/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.base;

import java.util.Comparator;
import java.util.HashMap;

import y.base.DataProvider;
import y.base.Edge;
import y.base.EdgeCursor;
import y.base.Graph;
import y.base.GraphEvent;
import y.base.GraphListener;
import y.base.ListCell;
import y.base.Node;
import y.base.NodeCursor;

/**
 * The base for the entire y25 package.
 * <p>
 * <h3>Layers</h3>
 * The <code>LayeredGraph</code> holds the graphs it is composed of in an
 * ordered list. Adding graphs to that list is either accomplished by creating
 * new graphs via the <code>createGraph</code> and <code>duplicateGraph</code>methods
 * or by importing an existing graph not created by the previous methods using
 * the <code>importGraph</code> method. The order of the graphs is guaranteed
 * not to change as long as no new graphs are created. Iterating over the graphs
 * is usually accomplished by getting a <code>GraphCursor</code> via the
 * <code>graphs</code> method. (See {@link GraphCursor GraphCursor} or the
 * package description for an example.)
 * <p>
 * <h3>LayerEdges</h3>
 * In addition to the graphs, the <code>LayeredGraph</code> also stores
 * <code>LayerEdges</code>. These are edges between arbitrary nodes inside
 * the layered graph, i.e. the layer edge can have a source node residing in one
 * graph and a target node residing in another graph - something that is not
 * allowed for ordinary edges as used in the yFiles library. Creating these
 * layer edges can be accomplished only by utilizing the layered graph
 * <code>createLayerEdge</code> method (since the only constructor of the
 * LayerEdge is made protected).
 * <p>
 * <h3>IDs</h3>
 * Most applications of the <code>LayeredGraph</code> require means of
 * identifying nodes and edges across multiple layers. Therefore, each node and
 * each edge inside a <code>LayeredGraph</code> has its own id. Newly created
 * nodes or edges automatically recieve a new and unique id (valid ids are in
 * the range <code>[1,inf)</code>), which can be read and changed by the
 * <code>get[Node|Edge]Id</code> and <code>set[Node|Edge]Id</code> methods.
 * Different nodes can have the same id, the same is true for edges. This way,
 * nodes or edges belonging to different layers inside the
 * <code>LayeredGraph</code> can be marked as beeing logically the same.
 * <p>
 * Additionaly, every node can be assigned a parent id. The default for every
 * node is a parent id of <code>0</code>, which means that this node has no
 * parents. Again, <code>getParentId</code> and <code>setParentId</code> can
 * be used to access the parent id of a node.
 * <p>
 * <h3>Maps</h3>
 * When working with a LayeredGraph, sometimes there is a need for attaching
 * data to its nodes, edges, layer edges or layers. For this reason, maps can be
 * created for each entities by using the
 * <code>create[Node|Edge|LayerEdge|Graph]Map</code> methods.
 * <p>
 * <h3>Building a <code>LayeredGraph</code></h3>
 * A common way to build a layered graph is to start with one layer, do some
 * work on it, then add another layer that has the same contents as the layer
 * before and modify some aspects of it, then create yet another layer with the
 * same contents as the second layer, do some modifications on it etc. This work
 * path can be accomplished by the following template:
 * 
 * <pre>
 *      // create the LayeredGraph
 *      LayeredGraph lg = new LayeredGraph();
 *      
 *      // create the first layer
 *      Graph currentLayer = lg.createGraph();
 *      
 *      // do some work on the first layer
 *      ...
 *      
 *      while ( stillNewLayersToAdd ) {
 *          
 *          // create a new layer with the same contents as the one before
 *          currentLayer = lg.duplicateGraph(currentLayer);
 *          
 *          // do some work on the new layer
 *          ...
 *      }
 * </pre>
 * 
 * <p>
 * <h3>Subclassing <code>LayeredGraph</code></h3>
 * When subclassing this layered graph, it is important to define the type of
 * graphs this layered graph will store. Currently, this layered graph stores
 * {@link y.base.Graph Graph}s, but subclasses may need to store other types,
 * for instance {@link y.layout.LayoutGraph LayoutGraph}s or
 * {@link y.view.Graph2D Graph2D}s.
 * <p>
 * Defining the type of graphs that are stored is done by assigning the type of
 * the used graphs to the variable {@link #graphType} in the constructor:
 * 
 * <pre>
 * graphType = Graph.class;
 * 
 * or
 * 
 * graphType = Graph2D.class;
 * </pre>
 * 
 * @author Werner Jainek
 * 
 */
public class LayeredGraph {

	/**
	 * Object insertion specifier. Object gets inserted before another one.
	 */
	public static final int BEFORE = 0;

	/**
	 * Object insertion specifier. Object gets inserted after another one.
	 */
	public static final int AFTER = 1;

	/**
	 * This key is used to register a DataProvider with a graph, that holds the
	 * IDs of the graphs nodes.
	 */
	public static final Object NODE_ID_DPKEY = new Object();

	/**
	 * This key is used to register a DataProvider with a graph, that holds the
	 * parent IDs of the graphs nodes.
	 */
	public static final Object PARENT_ID_DPKEY = new Object();

	/**
	 * This key is used to register a DataProvider with a graph, that holds the
	 * IDs of the graphs edges.
	 */
	public static final Object EDGE_ID_DPKEY = new Object();

	/**
	 * Specifies the type of the graphs this layered graph stores in the
	 * graphList. This type is also used whenever a new Graph has to be created,
	 * for example in the <code>createGraph</code> methods. Therefore the type
	 * stored in this variable always needs to have a default constructor !
	 */
	protected Class graphType;

	/**
	 * Stores the graphs that compose this layered graph in an ordered fashion.
	 */
	protected GraphList graphList;

	/**
	 * Stores all edges between different layers in this layered graph.
	 */
	protected LayerEdgeList layerEdgeList;

	/**
	 * Stores a list of outgoing layer edges for each node that has at least one
	 * outgoing layer edge.
	 */
	protected NodeMap outLayerEdges;

	/**
	 * Stores a list of incoming layer edges for each node that has at least one
	 * incoming layer edge.
	 */
	protected NodeMap inLayerEdges;

	/**
	 * Assigns new and unique IDs to all newly created nodes and edges.
	 */
	protected IDAssigner idAssigner;

	/**
	 * Utilized by <code>idAssigner</code> to create new node IDs and to store
	 * and free already used IDs.
	 */
	protected IDCreator nodeIDCreator;

	/**
	 * Utilized by <code>idAssigner</code> to create new edge IDs and to store
	 * and free already used IDs.
	 */
	protected IDCreator edgeIDCreator;

	// CONSTRUCTOR / RESETTING --------------------------------------------

	/**
	 * Instantiates an empty LayeredGraph object.
	 */
	public LayeredGraph() {
		graphType = Graph.class;
		graphList = new GraphList();
		layerEdgeList = new LayerEdgeList();
		inLayerEdges = createNodeMap();
		outLayerEdges = createNodeMap();
		idAssigner = new IDAssigner();
		nodeIDCreator = new IDCreator();
		edgeIDCreator = new IDCreator();
	}

	// TODO: Implement a clone() method !
	// public LayeredGraph( LayeredGraph g );

	/**
	 * Resets the LayeredGraph to its initial state. More precisely, this means
	 * that all graphs and layer edges are removed and the automatic ID
	 * assignment process is reset (i.e. newly created IDs start from 1).
	 * <p>
	 * Subclasses should override this method by first doing the work necessary
	 * on their part and then calling <code>clear</code>.
	 */
	public void clear() {
		graphList = new GraphList();

		layerEdgeList = new LayerEdgeList();
		inLayerEdges = createNodeMap();
		outLayerEdges = createNodeMap();

		idAssigner = new IDAssigner();
		nodeIDCreator = new IDCreator();
		edgeIDCreator = new IDCreator();
	}

	/**
	 * Removes all layer edges present in this LayeredGraph.
	 * <p>
	 * Subclasses should override this method by first doing the work necessary
	 * on their part and then calling <code>super.clearLayerEdges</code>.
	 */
	public void clearLayerEdges() {
		layerEdgeList = new LayerEdgeList();
		inLayerEdges = createNodeMap();
		outLayerEdges = createNodeMap();
	}

	// CREATING / REMOVING GRAPHS
	// ------------------------------------------------

	/**
	 * Creates a new graph and returns it. The new graph will be located at the
	 * end of the graph list inside the LayeredGraph.
	 * <p>
	 * The returned graph automatically assigns new and unique IDs to nodes and
	 * edges as soon as they are created. After creation, changing these
	 * automatically assigned IDs is accomplished by calling the LayeredGraph
	 * methods <code>setNodeID</code>, <code>setParentID</code> and
	 * <code>setEdgeID</code>.
	 * 
	 * @return The newly created graph.
	 */
	public Graph createGraph() {
		return createGraph(AFTER, null);
	}

	/**
	 * Creates a new graph at a specified location and returns it.
	 * <p>
	 * The returned graph automatically assigns new and unique IDs to nodes and
	 * edges as soon as they are created. After creation, changing these
	 * automatically assigned IDs is accomplished by calling the LayeredGraph
	 * methods <code>setNodeID</code>, <code>setParentID</code> and
	 * <code>setEdgeID</code>.
	 * 
	 * @param where
	 *            Can be either <code>LayeredGraph.BEFORE</code> or
	 *            <code>LayeredGraph.AFTER</code>. In the fromer case, the
	 *            new graph is inserted into the list right before the passed
	 *            reference graph, in the latter case it is inserted after the
	 *            passed reference graph.
	 * @param reference
	 *            A graph that is already in the graph list of this
	 *            LayeredGraph. If <code>null</code> is passed, the newly
	 *            created graph is inserted at the very beginning of the list if
	 *            <code>where</code> is <code>LayeredGraph.BEFORE</code> or
	 *            at the very end of the graph list if <code>where</code> is
	 *            <code>LayeredGraph.AFTER</code>.
	 * 
	 * @return The newly created graph.
	 */
	public Graph createGraph(int where, Graph reference) {

		// this complicated code is necessary so that subclasses can just assign
		// another type to graphType and - voila - the correct type is
		// automatically created !
		Graph g = null;
		try {
			g = (Graph) graphType.newInstance();
		} catch (InstantiationException e) {
		} catch (IllegalAccessException e) {
		}

		// create the necessary structure data
		g.addDataProvider(NODE_ID_DPKEY, g.createNodeMap());
		g.addDataProvider(PARENT_ID_DPKEY, g.createNodeMap());
		g.addDataProvider(EDGE_ID_DPKEY, g.createEdgeMap());

		// register graph listener that attaches appropriate id's to newly
		// created nodes/edges
		g.addGraphListener(idAssigner);

		// add graph to list
		if (reference != null) {
			switch (where) {
			default:
			case AFTER:
				graphList.insertAfter(g, graphList.findCell(reference));
				break;
			case BEFORE:
				graphList.insertBefore(g, graphList.findCell(reference));
				break;
			}
		} else {
			switch (where) {
			default:
			case AFTER:
				graphList.addLast(g);
				break;
			case BEFORE:
				graphList.addFirst(g);
				break;
			}
		}

		return g;
	}

	/**
	 * Creates a new graph by copying the passed graph. The passed graph has to
	 * be part of this <code>LayeredGraph</code>. The new graph's location
	 * inside the graph list will be right after the original graph.
	 * <p>
	 * The new graph is then returned. It has exactly the same nodes and edges
	 * as the original graph. All node, edge and parent IDs are also identical
	 * to the original. The returned graph automatically assigns new and unique
	 * IDs to newly created nodes and edges. After creation, changing these
	 * automatically assigned IDs is accomplished by calling the LayeredGraph
	 * methods <code>setNodeID</code>, <code>setParentID</code> and
	 * <code>setEdgeID</code>.
	 * 
	 * @param original
	 *            the graph that will be copied.
	 * @return the newly created graph that is a duplicate of the original.
	 */
	public Graph duplicateGraph(Graph original) {

		// create information about which datamaps to copy from the original
		// graph
		GraphTools.GraphDataInfo dataInfo[] = {
				new GraphTools.GraphDataInfo(NODE_ID_DPKEY,
						GraphTools.GraphDataInfo.NODEMAP,
						GraphTools.GraphDataInfo.INTEGER),
				new GraphTools.GraphDataInfo(PARENT_ID_DPKEY,
						GraphTools.GraphDataInfo.NODEMAP,
						GraphTools.GraphDataInfo.INTEGER),
				new GraphTools.GraphDataInfo(EDGE_ID_DPKEY,
						GraphTools.GraphDataInfo.EDGEMAP,
						GraphTools.GraphDataInfo.INTEGER) };

		// create a copy of the given graph
		Graph copy = GraphTools.DuplicateGraph(original, dataInfo);

		// register graph listener that attaches appropriate id's to newly
		// created nodes/edges
		copy.addGraphListener(idAssigner);

		// insert copy into graph list
		graphList.insertAfter(copy, graphList.findCell(original));

		return copy;
	}

	/**
	 * Creates a new graph by copying the passed graph. The difference to the
	 * <code>duplicateGraph</code> method is, that the passed graph is not
	 * part of this layered graph. If you already have code that generates a
	 * graph and you want to use that graph inside a layered graph, this is the
	 * method you have to use. (It is an error to pass a graph that is already
	 * inside this layered graph to this method.)
	 * <p>
	 * The new graph will be located at the end of the graph list inside the
	 * LayeredGraph. It has exactly the same nodes and edges as the original
	 * graph.
	 * <p>
	 * Regarding node, edge and parent IDs, there are two possibilities:
	 * <ol>
	 * <li> If the graph you pass has no <code>DataProvider</code>s for the
	 * keys <code>LayeredGraph.NODE_ID_DPKEY</code>,
	 * <code>LayeredGraph.PARENT_ID_DPKEY</code> and
	 * <code>LayeredGraph.EDGE_ID_DPKEY</code>, new IDs will be created for
	 * every node and edge in the copied graph. The parent IDs will all be set
	 * to <code>0</code>. </li>
	 * <li> If you want to specify node, edge and parent IDs, you have to
	 * register <code>DataProvider</code>s for the following keys:
	 * <code>LayeredGraph.NODE_ID_DPKEY</code>,
	 * <code>LayeredGraph.PARENT_ID_DPKEY</code>,
	 * <code>LayeredGraph.EDGE_ID_DPKEY</code>. Then you can assign node,
	 * edge and parent IDs as you like. These IDs will then be copied to the new
	 * graph that is created. Note that you have to perform these steps
	 * <em>before</em> you call this method. </li>
	 * </ol>
	 * The new graph automatically assigns new and unique IDs to newly created
	 * nodes and edges. After creation, changing these automatically assigned
	 * IDs is accomplished by calling the LayeredGraph methods
	 * <code>setNodeID</code>, <code>setParentID</code> and
	 * <code>setEdgeID</code>.
	 * 
	 * @param graph
	 *            the graph that will be copied.
	 * @return the newly created graph that is a duplicate of the passed graph.
	 */
	public Graph importGraph(Graph graph) {

		// copy the graph without the structure information
		Graph copy = GraphTools.DuplicateGraph(graph, null);

		// copy or create the necessary structure data
		copyCreateStructureData(graph, copy);

		// register graph listener that attaches appropriate id's to newly
		// created nodes/edges
		copy.addGraphListener(idAssigner);

		// insert copy into graph list
		graphList.addLast(copy);

		return copy;
	}

	/**
	 * Helper method for <code>importGraph</code>.
	 * <p>
	 * This method copies structure data (node, edge and parent IDs) from
	 * <code>graph</code> to <code>copy</code>. <code>copy</code> has to
	 * have exactly the same number and order of nodes and edges as
	 * <code>graph</code>.
	 * <p>
	 * If the original graph has only partial structure information (say, it
	 * only has node IDs but no edge and no parent IDs, and not every node has a
	 * valid ID (!=0)), structure data is created as follows: Nodes without IDs
	 * get a new ID, edges without IDs get a new ID and nodes without a parent
	 * ID get <code>0</code> assigned.
	 * 
	 * @param graph
	 * @param copy
	 */
	protected void copyCreateStructureData(Graph graph, Graph copy) {

		// find out which maps are already existant in graph
		boolean hasNodeID = false;
		boolean hasParentID = false;
		boolean hasEdgeID = false;

		y.base.NodeMap graphNodeIDMap = null;
		y.base.NodeMap graphParentIDMap = null;
		y.base.EdgeMap graphEdgeIDMap = null;

		Object keys[] = graph.getDataProviderKeys();
		for (int i = 0; i < keys.length; ++i) {
			if (keys[i] == NODE_ID_DPKEY) {
				hasNodeID = true;
				graphNodeIDMap = (y.base.NodeMap) graph
						.getDataProvider(keys[i]);
			} else if (keys[i] == PARENT_ID_DPKEY) {
				hasParentID = true;
				graphParentIDMap = (y.base.NodeMap) graph
						.getDataProvider(keys[i]);
			} else if (keys[i] == EDGE_ID_DPKEY) {
				hasEdgeID = true;
				graphEdgeIDMap = (y.base.EdgeMap) graph
						.getDataProvider(keys[i]);
			}
		}

		// create all maps in the copy
		y.base.NodeMap copyNodeIDMap = copy.createNodeMap();
		y.base.NodeMap copyParentIDMap = copy.createNodeMap();
		y.base.EdgeMap copyEdgeIDMap = copy.createEdgeMap();

		// copy/create --- NODE ID ---
		if (!hasNodeID) {
			for (NodeCursor nc = copy.nodes(); nc.ok(); nc.next()) {
				copyNodeIDMap.setInt(nc.node(), nodeIDCreator.createID());
			}
		} else {
			NodeCursor gNC = graph.nodes();
			NodeCursor cNC = copy.nodes();

			while (cNC.ok() && gNC.ok()) {
				int gID = graphNodeIDMap.getInt(gNC.node());
				int cID = (gID == 0) ? nodeIDCreator.createID() : gID;

				copyNodeIDMap.setInt(cNC.node(), cID);

				gNC.next();
				cNC.next();
			}
		}

		// copy/create --- PARENT ID ---
		if (!hasParentID) {
			for (NodeCursor nc = copy.nodes(); nc.ok(); nc.next()) {
				copyParentIDMap.setInt(nc.node(), 0);
			}
		} else {
			NodeCursor gNC = graph.nodes();
			NodeCursor cNC = copy.nodes();

			while (cNC.ok() && gNC.ok()) {
				int gID = graphParentIDMap.getInt(gNC.node());
				int cID = gID; // if gID is 0, then just use that

				copyParentIDMap.setInt(cNC.node(), cID);

				gNC.next();
				cNC.next();
			}
		}

		// copy/create --- EDGE ID ---
		if (!hasEdgeID) {
			for (EdgeCursor ec = copy.edges(); ec.ok(); ec.next()) {
				copyEdgeIDMap.setInt(ec.edge(), edgeIDCreator.createID());
			}
		} else {
			EdgeCursor gEC = graph.edges();
			EdgeCursor cEC = copy.edges();

			while (cEC.ok() && gEC.ok()) {
				int gID = graphEdgeIDMap.getInt(gEC.edge());
				int cID = (gID == 0) ? edgeIDCreator.createID() : gID;

				copyEdgeIDMap.setInt(cEC.edge(), cID);

				gEC.next();
				cEC.next();
			}
		}

		// register data providers
		copy.addDataProvider(NODE_ID_DPKEY, copyNodeIDMap);
		copy.addDataProvider(PARENT_ID_DPKEY, copyParentIDMap);
		copy.addDataProvider(EDGE_ID_DPKEY, copyEdgeIDMap);

		// record all used IDs
		y.base.NodeMap nm = (y.base.NodeMap) copy
				.getDataProvider(NODE_ID_DPKEY);
		for (NodeCursor nc = copy.nodes(); nc.ok(); nc.next()) {
			nodeIDCreator.setUsed(nm.getInt(nc.node()));
		}

		y.base.EdgeMap em = (y.base.EdgeMap) copy
				.getDataProvider(EDGE_ID_DPKEY);
		for (EdgeCursor ec = copy.edges(); ec.ok(); ec.next()) {
			edgeIDCreator.setUsed(em.getInt(ec.edge()));
		}
	}

	/**
	 * Removes the passed graph from this LayeredGraph.
	 * <p>
	 * All node and edge IDs that were associated with some node or edge inside
	 * this graph will be freed for reuse by the automatic ID assignment
	 * process. This is only the case, however, if the same node or edge ID is
	 * not used by another node or edge inside this LayeredGraph. If a node ID
	 * is freed, the parent IDs of all child nodes are set to <code>0</code>.
	 * <p>
	 * All layer edges originating or ending at some node inside the removed
	 * graph will be removed as well.
	 * <p>
	 * Subclasses should override this method by first doing the work necessary
	 * on their part and then calling <code>super.removeGraph</code>.
	 * <p>
	 * After the graph has been removed, it can still be used by the user. The
	 * LayeredGraph has no further responsibility for the graph and the graph is
	 * not tied to the LayeredGraph by any means. The structure information that
	 * was present before removal (node/parent/edge ids) is not deleted,
	 * however, and is still available in the removed graph. To retrieve this
	 * information, no mehods from LayeredGraph can be used, but one has to
	 * manually retrieve the DataProviders for the given keys (<code>LayeredGraph.NODE_ID_DPKEY</code>,
	 * <code>LayeredGraph.PARENT_ID_DPKEY</code>,
	 * <code>LayeredGraph.EDGE_ID_DPKEY</code>) instead. Since all ties to
	 * the LayeredGraph are broken, newly created nodes and edges in the removed
	 * graph do not automatically recieve a new ID. Also, creation of new layer
	 * edges starting or ending in that graph is no longer possible.
	 * 
	 * @param graph
	 *            The graph that will be removed.
	 */
	public void removeGraph(Graph graph) {

		// first, we remove the graph from the list
		graphList.remove(graph);

		// then, we free all IDs that were used only by this graph
		idAssigner.removeIDs(graph);
		graph.removeGraphListener(idAssigner);

		// finally, we remove all layer edges originating from or ending at some
		// node inside this graph
		for (NodeCursor nc = graph.nodes(); nc.ok(); nc.next()) {
			Node n = nc.node();

			// remove all in layer edges
			for (LayerEdgeCursor lec = inLayerEdges(n); lec.ok(); lec.next()) {
				removeLayerEdge(lec.layerEdge()); // this works, because the
				// cursors don't become
				// invalid if the list cell
				// they point to gets
				// removed as long as it is
				// not added to another list
				// (see YList documentation)
			}

			// remove all out layer edges for that node
			for (LayerEdgeCursor lec = outLayerEdges(n); lec.ok(); lec.next()) {
				removeLayerEdge(lec.layerEdge());
			}
		}
	}

	// CREATING / REMOVING LAYER EDGES
	// ------------------------------------------------

	/**
	 * Creates a new layer edge from the source to the target node. Both nodes
	 * must be inside graphs that belong to this LayeredGraph.
	 * 
	 * @param source
	 *            The source node for the layer edge.
	 * @param target
	 *            The target node for the layer edge.
	 * @return The newly created layer edge.
	 */
	public LayerEdge createLayerEdge(Node source, Node target) {

		// A. Create a new layer edge and add it to the list
		LayerEdge le = new LayerEdge(this, source, target);
		layerEdgeList.add(le);

		// B. Add it to the outgoing and incoming layer edge lists of the source
		// and target nodes

		// B-1. source node
		LayerEdgeList sourceOutList = (LayerEdgeList) outLayerEdges.get(source);

		// if this is the first outgoing layer edge, we have to create the list
		if (sourceOutList == null) {
			sourceOutList = new LayerEdgeList();
			outLayerEdges.set(source, sourceOutList);
		}

		sourceOutList.add(le);

		// B-2. target node
		LayerEdgeList targetInList = (LayerEdgeList) inLayerEdges.get(target);

		// if this is the first outgoing layer edge, we have to create the list
		if (targetInList == null) {
			targetInList = new LayerEdgeList();
			inLayerEdges.set(target, targetInList);
		}

		targetInList.add(le);

		// C. finally, return the layer edge
		return le;
	}

	/**
	 * Removes the passed layer edge from this LayeredGraph.
	 * <p>
	 * Subclasses should override this method by first doing the work necessary
	 * on their part and then calling <code>super.removeLayerEdge</code>.
	 * 
	 * @param layerEdge
	 *            The layer edge that will be removed.
	 */
	public void removeLayerEdge(LayerEdge layerEdge) {

		layerEdgeList.remove(layerEdge);

		LayerEdgeList sourceOutList = (LayerEdgeList) outLayerEdges
				.get(layerEdge.source());
		sourceOutList.remove(layerEdge);

		LayerEdgeList targetInList = (LayerEdgeList) inLayerEdges.get(layerEdge
				.target());
		targetInList.remove(layerEdge);

		// if the removed layer edge was the only layer edge for the source or
		// target node, we remove the list as well
		if (sourceOutList.isEmpty())
			outLayerEdges.set(layerEdge.source(), null);
		if (targetInList.isEmpty())
			inLayerEdges.set(layerEdge.target(), null);
	}

	// CHANGING GRAPH ORDER
	// ------------------------------------------------
	/**
	 * Moves the passed graph forward in the graph list. The passed graph has to
	 * be part of this layered graph. If it is already at the front of the graph
	 * list, nothing happens.
	 * <p>
	 * The first entry in the graph list is considered to be the backmost, the
	 * last entry is considered to be frontmost.
	 */
	public void moveForward(Graph graph) {
		ListCell cell = graphList.findCell(graph);
		ListCell next = cell.succ();

		// maybe graph is already at front?
		if (next == null) {
			return;
		}

		// insert graph in list cell after next
		graphList.insertAfter(graph, next);

		// remove the old list cell
		graphList.removeCell(cell);
	}

	/**
	 * Moves the passed graph to the front of the graph list. The passed graph
	 * has to be part of this layered graph. If it is already at the front of
	 * the graph list, nothing happens.
	 * <p>
	 * The first entry in the graph list is considered to be the backmost, the
	 * last entry is considered to be frontmost.
	 */
	public void moveToFront(Graph graph) {
		ListCell cell = graphList.findCell(graph);
		ListCell frontCell = graphList.lastCell();

		// maybe graph is already at front?
		if (cell == frontCell) {
			return;
		}

		// insert graph in list cell after frontCell
		graphList.insertAfter(graph, frontCell);

		// remove the old list cell
		graphList.removeCell(cell);
	}

	/**
	 * Moves the passed graph backward in the graph list. The passed graph has
	 * to be part of this layered graph. If it is already at the back of the
	 * graph list, nothing happens.
	 * <p>
	 * The first entry in the graph list is considered to be the backmost, the
	 * last entry is considered to be frontmost.
	 */
	public void moveBackward(Graph graph) {
		ListCell cell = graphList.findCell(graph);
		ListCell prev = cell.pred();

		// maybe graph is already at back?
		if (prev == null) {
			return;
		}

		// insert graph in list cell before prev
		graphList.insertBefore(graph, prev);

		// remove the old list cell
		graphList.removeCell(cell);
	}

	/**
	 * Moves the passed graph to the back of the graph list. The passed graph
	 * has to be part of this layered graph. If it is already at the back of the
	 * graph list, nothing happens.
	 * <p>
	 * The first entry in the graph list is considered to be the backmost, the
	 * last entry is considered to be frontmost.
	 */
	public void moveToBack(Graph graph) {
		ListCell cell = graphList.findCell(graph);
		ListCell backCell = graphList.firstCell();

		// maybe graph is already at back?
		if (cell == backCell) {
			return;
		}

		// insert graph in list cell before backCell
		graphList.insertBefore(graph, backCell);

		// remove the old list cell
		graphList.removeCell(cell);
	}

	// ASSIGNING NODE / PARENT / EDGE IDS
	// ------------------------------------------------

	/**
	 * Returns the ID of the passed node. Valid node IDs are in the range
	 * <code>[1,inf)</code>.
	 */
	public int getNodeID(Node n) {
		return n.getGraph().getDataProvider(NODE_ID_DPKEY).getInt(n);
	}

	/**
	 * Returns the ID of the parent of the passed node. A return value of
	 * <code>0</code> indicates, that the passed node has no parent.
	 */
	public int getParentID(Node n) {
		return n.getGraph().getDataProvider(PARENT_ID_DPKEY).getInt(n);
	}

	/**
	 * Returns the edge ID of the passed edge. Valid edge IDs are in the range
	 * <code>[1,inf)</code>.
	 */
	public int getEdgeID(Edge e) {
		return e.getGraph().getDataProvider(EDGE_ID_DPKEY).getInt(e);
	}

	/**
	 * Sets the ID of the passed node to the value specified by <code>id</code>.
	 * 
	 * @param n
	 *            The node whose ID is to be set.
	 * @param id
	 *            The new ID of the node. Valid IDs are in the range
	 *            <code>[1,inf)</code>. When <code>0</code> is passed, a
	 *            new and unique ID will be created that has not been used by
	 *            any other node. If a value less than <code>0</code> is
	 *            passed, this method will not fail but assign that value to the
	 *            node even if it is not a valid id. This way you can
	 *            temporarily mark nodes, but you are responsible to reset the
	 *            ID to a valid number.
	 * @return The ID that has been assigned to the given node.
	 */
	public int setNodeID(Node n, int id) {
		// TODO: check old ID and remove it if it doesnt occur again ?
		int i = (id == 0) ? nodeIDCreator.createID() : id;
		if (i > 0)
			nodeIDCreator.setUsed(i);
		((y.base.NodeMap) (n.getGraph().getDataProvider(NODE_ID_DPKEY)))
				.setInt(n, i);
		return i;
	}

	/**
	 * Sets the parent ID for the given node.
	 * 
	 * @param n
	 *            The node whose parent ID is to be set.
	 * @param id
	 *            The ID of the parent of the given node. Valid IDs are in the
	 *            range <code>[1,inf)</code>. A value of <code>0</code> is
	 *            also allowed and means, that the given node has no parent.
	 *            When a value less than <code>0</code> is passed the method
	 *            will not fail but stores that value as the parent ID of the
	 *            node even if it is not a valid ID. This way you can
	 *            temporarily mark nodes, but you are responsible to reset the
	 *            ID to a valid number. Note that no actual node with the passed
	 *            ID needs to exist. This gives you the possibility of assigning
	 *            a parent ID to a node even if the parent node itself is not
	 *            created yet. To ensure that this works correctly, the
	 *            automatic ID assignment process is notified of the new parent
	 *            ID, so that it does not accidentaly assign that ID to a newly
	 *            created node.
	 * @return The parent ID that has been assigned to the given node.
	 */
	public int setParentID(Node n, int id) {
		if (id > 0)
			nodeIDCreator.setUsed(id);
		((y.base.NodeMap) n.getGraph().getDataProvider(PARENT_ID_DPKEY))
				.setInt(n, id);
		return id;
	}

	/**
	 * Sets the ID of the passed edge to the value specified by <code>id</code>.
	 * 
	 * @param e
	 *            The edge whose ID is to be set.
	 * @param id
	 *            The new ID of the edge. Valid IDs are in the range
	 *            <code>[1,inf)</code>. When <code>0</code> is passed, a
	 *            new and unique ID will be created that has not been used by
	 *            any other edge. If a value less than <code>0</code> is
	 *            passed, this method will not fail but assign that value to the
	 *            edge even if it is not a valid id. This way you can
	 *            temporarily mark edges, but you are responsible to reset the
	 *            ID to a valid number.
	 * @return The ID that has been assigned to the given edge.
	 */
	public int setEdgeID(Edge e, int id) {
		int i = (id == 0) ? edgeIDCreator.createID() : id;
		if (id > 0)
			edgeIDCreator.setUsed(i);
		((y.base.EdgeMap) e.getGraph().getDataProvider(EDGE_ID_DPKEY)).setInt(
				e, i);
		return i;
	}

	// GRAPH INFORMATION ------------------------------------------------
	/**
	 * Returns the number of graphs inside this LayeredGraph.
	 */
	public int G() {
		return graphList.size();
	}

	/**
	 * Returns the number of graphs inside this LayeredGraph.
	 */
	public int graphCount() {
		return G();
	}

	/**
	 * Returns the number of layer edges inside this LayeredGraph.
	 */
	public int LE() {
		return layerEdgeList.size();
	}

	/**
	 * Returns the number of layer edges inside this LayeredGraph.
	 */
	public int layerEdgeCount() {
		return LE();
	}

	/**
	 * Returns the number of nodes inside this LayeredGraph. This is just the
	 * sum of the node counts of the graphs that belong to this LayeredGraph.
	 */
	public int N() {
		int c = 0;
		for (GraphCursor gc = graphs(); gc.ok(); gc.next()) {
			c += gc.graph().N();
		}
		return c;
	}

	/**
	 * Returns the number of nodes inside this LayeredGraph. This is just the
	 * sum of the node counts of the graphs that belong to this LayeredGraph.
	 */
	public int nodeCount() {
		return N();
	}

	/**
	 * Returns the number of edges (layer edges are not included) inside this
	 * LayeredGraph. This is just the sum of the edge counts of the graphs that
	 * belong to this LayeredGraph.
	 */
	public int E() {
		int c = 0;
		for (GraphCursor gc = graphs(); gc.ok(); gc.next()) {
			c += gc.graph().E();
		}
		return c;
	}

	/**
	 * Returns the number of edges (layer edges are not included) inside this
	 * LayeredGraph. This is just the sum of the edge counts of the graphs that
	 * belong to this LayeredGraph.
	 */
	public int edgeCount() {
		return E();
	}

	// ITERATION MECHANISMS / ORDER
	// ------------------------------------------------

	/**
	 * Sorts the graphs with respect to the given <code>Comparator</code>.
	 */
	public void sortGraphs(Comparator comparator) {
		graphList.sort(comparator);
	}

	/**
	 * Returns a cursor over all graphs inside this LayeredGraph.
	 */
	public GraphCursor graphs() {
		return graphList.graphs();
	}

	/**
	 * Returns a cursor over all graphs inside this LayeredGraph that occur
	 * after the given reference. So, if the list of graphs can be decomposed
	 * into 3 sublists [g_first...g_(ref-1)] [g_ref] [g_(ref+1)...g_last], this
	 * method returns a cursor over the last sublist.
	 * 
	 * @param ref
	 *            The reference graph as described above.
	 */
	public GraphCursor getGraphCursorAfter(Graph ref) {
		GraphCursor gc = graphs();
		for (; gc.ok(); gc.next()) {
			if (gc.graph() == ref) {
				gc.next();
				return gc;
			}
		}
		return gc;
	}

	/**
	 * Returns a cursor over all graphs inside this LayeredGraph that occur
	 * before the given reference. So, if the list of graphs can be decomposed
	 * into 3 sublists [g_first...g_(ref-1)] [g_ref] [g_(ref+1)...g_last], this
	 * method returns a cursor over the first sublist.
	 * 
	 * @param ref
	 *            The reference graph as described above.
	 */
	public GraphCursor getGraphCursorBefore(Graph ref) {
		GraphCursor gc = graphs();
		gc.toLast();
		for (; gc.ok(); gc.prev()) {
			if (gc.graph() == ref) {
				gc.prev();
				return gc;
			}
		}
		return gc;
	}

	/**
	 * Returns a cursor over all layer edges inside this LayeredGraph.
	 */
	public LayerEdgeCursor layerEdges() {
		return layerEdgeList.layerEdges();
	}

	/**
	 * Returns a cursor over all layer edges that originate from the given node.
	 * 
	 * @param node
	 *            The source of the layer edges.
	 * @return A cursor over all outgoing layer edges starting at the given
	 *         node.
	 */
	public LayerEdgeCursor outLayerEdges(Node node) {
		LayerEdgeList nodeOutList = (LayerEdgeList) outLayerEdges.get(node);

		// if there are no out layer edges, we create an empty list so that we
		// can return a cursor, but we don't do anything else (e.g storing the
		// newly creates list). So when the LayerEdgeCursor is not used anymore,
		// the created list gets garbage collected as well
		if (nodeOutList == null) {
			nodeOutList = new LayerEdgeList();
		}

		return nodeOutList.layerEdges();
	}

	/**
	 * Returns a cursor over all layer edges that have the given node as their
	 * target.
	 * 
	 * @param node
	 *            The target of the layer edges.
	 * @return A cursor over all incoming layer edges at the given node.
	 */
	public LayerEdgeCursor inLayerEdges(Node node) {
		LayerEdgeList nodeInList = (LayerEdgeList) inLayerEdges.get(node);

		// if there are no in layer edges, we create an empty list so that we
		// can return a cursor, but we don't do anything else (e.g storing the
		// newly creates list). So when the LayerEdgeCursor is not used anymore,
		// the created list gets garbage collected as well
		if (nodeInList == null) {
			nodeInList = new LayerEdgeList();
		}

		return nodeInList.layerEdges();
	}

	/**
	 * Returns an array of all graphs inside this LayeredGraph.
	 */
	public Graph[] getGraphArray() {
		Graph garray[] = new Graph[graphList.size()];
		int i = 0;
		for (GraphCursor gc = graphs(); gc.ok(); gc.next()) {
			garray[i] = gc.graph();
			++i;
		}
		return garray;
	}

	/**
	 * Returns an array of all layer edges inside this LayeredGraph.
	 */
	public LayerEdge[] getLayerEdgeArray() {
		LayerEdge learray[] = new LayerEdge[layerEdgeList.size()];
		int i = 0;
		for (LayerEdgeCursor lec = layerEdges(); lec.ok(); lec.next()) {
			learray[i] = lec.layerEdge();
			++i;
		}
		return learray;
	}

	// ----------------------------------------------------------------------

	// ADDING EXISTING GRAPHS ------------------------------------------------

	// NODE, EDGE, GRAPH and LAYER EDGE MAPS
	// ---------------------------------------

	/**
	 * Creates a <code>NodeMap</code> and returns it. In contrast to the
	 * {@link y.base.NodeMap y.base.NodeMap} this <code>NodeMap</code> can be
	 * used for mapping values to nodes from different graphs.
	 */
	public NodeMap createNodeMap() {
		return new NodeMapImpl();
	}

	/**
	 * Creates an <code>EdgeMap</code> and returns it. In contrast to the
	 * {@link y.base.EdgeMap y.base.EdgeMap} this <code>EdgeMap</code> can be
	 * used for mapping values to edges from different graphs.
	 */
	public EdgeMap createEdgeMap() {
		return new EdgeMapImpl();
	}

	/**
	 * Creates a <code>GraphMap</code> and returns it.
	 */
	public GraphMap createGraphMap() {
		return new GraphMapImpl();

	}

	/**
	 * Creates a <code>LayerEdgeMap</code> and returns it.
	 */
	public LayerEdgeMap createLayerEdgeMap() {
		return new LayerEdgeMapImpl();
	}

	/**
	 * An Implementation of the <code>NodeMap</code> interface based on a
	 * <code>HashMap</code>.
	 * 
	 * @author Werner Jainek
	 */
	protected class NodeMapImpl implements NodeMap {

		/**
		 * The <code>HashMap</code> that is used to store key/value pairs.
		 */
		protected HashMap map;

		NodeMapImpl() {
			map = new HashMap();
		}

		public Object remove(Object node) {
			return map.remove(node);
		}

		public void set(Object key, Object value) {
			map.put(key, value);
		}

		public Object get(Object key) {
			return map.get(key);
		}

		public boolean getBool(Object node) {
			Object o = get(node);
			if (o == null)
				return false;
			else
				return ((Boolean) o).booleanValue();
		}

		public double getDouble(Object node) {
			Object o = get(node);
			if (o == null)
				return 0.0;
			else
				return ((Double) o).doubleValue();
		}

		public int getInt(Object node) {
			Object o = get(node);
			if (o == null)
				return 0;
			else
				return ((Integer) o).intValue();
		}

		public void setBool(Object node, boolean value) {
			set(node, new Boolean(value));
		}

		public void setDouble(Object node, double value) {
			set(node, new Double(value));
		}

		public void setInt(Object node, int value) {
			set(node, new Integer(value));
		}
	}

	/**
	 * An Implementation of the <code>EdgeMap</code> interface based on a
	 * <code>HashMap</code>.
	 * 
	 * @author Werner Jainek
	 */
	protected class EdgeMapImpl implements EdgeMap {

		/**
		 * The <code>HashMap</code> that is used to store key/value pairs.
		 */
		protected HashMap map;

		EdgeMapImpl() {
			map = new HashMap();
		}

		public Object remove(Object edge) {
			return map.remove(edge);
		}

		public void set(Object key, Object value) {
			map.put(key, value);
		}

		public Object get(Object key) {
			return map.get(key);
		}

		public boolean getBool(Object edge) {
			Object o = get(edge);
			if (o == null)
				return false;
			else
				return ((Boolean) o).booleanValue();
		}

		public double getDouble(Object edge) {
			Object o = get(edge);
			if (o == null)
				return 0.0;
			else
				return ((Double) o).doubleValue();
		}

		public int getInt(Object edge) {
			Object o = get(edge);
			if (o == null)
				return 0;
			else
				return ((Integer) o).intValue();
		}

		public void setBool(Object edge, boolean value) {
			set(edge, new Boolean(value));
		}

		public void setDouble(Object edge, double value) {
			set(edge, new Double(value));
		}

		public void setInt(Object edge, int value) {
			set(edge, new Integer(value));
		}
	}

	/**
	 * An Implementation of the <code>LayerEdgeMap</code> interface based on a
	 * <code>HashMap</code>.
	 * 
	 * @author werner
	 */
	protected class LayerEdgeMapImpl implements LayerEdgeMap {

		/**
		 * The <code>HashMap</code> that is used to store key/value pairs.
		 */
		protected HashMap map;

		public LayerEdgeMapImpl() {
			map = new HashMap();
		}

		public Object remove(Object layerEdge) {
			return map.remove(layerEdge);
		}

		public Object get(Object layerEdge) {
			return map.get(layerEdge);
		}

		public void set(Object layerEdge, Object value) {
			map.put(layerEdge, value);
		}

		public boolean getBool(Object layerEdge) {
			Object o = get(layerEdge);
			if (o == null)
				return false;
			else
				return ((Boolean) o).booleanValue();
		}

		public double getDouble(Object layerEdge) {
			Object o = get(layerEdge);
			if (o == null)
				return 0.0;
			else
				return ((Double) o).doubleValue();
		}

		public int getInt(Object layerEdge) {
			Object o = get(layerEdge);
			if (o == null)
				return 0;
			else
				return ((Integer) o).intValue();
		}

		public void setBool(Object layerEdge, boolean value) {
			set(layerEdge, new Boolean(value));
		}

		public void setDouble(Object layerEdge, double value) {
			set(layerEdge, new Double(value));
		}

		public void setInt(Object layerEdge, int value) {
			set(layerEdge, new Integer(value));
		}

	}

	/**
	 * An Implementation of the <code>GraphMap</code> interface based on a
	 * <code>HashMap</code>.
	 * 
	 * @author werner
	 */
	protected class GraphMapImpl implements GraphMap {

		/**
		 * The <code>HashMap</code> that is used to store key/value pairs.
		 */
		protected HashMap map;

		GraphMapImpl() {
			map = new HashMap();
		}

		public Object remove(Object graph) {
			return map.remove(graph);
		}

		public void set(Object key, Object value) {
			map.put(key, value);
		}

		public Object get(Object key) {
			return map.get(key);
		}

		public boolean getBool(Object graph) {
			Object o = get(graph);
			if (o == null)
				return false;
			else
				return ((Boolean) o).booleanValue();
		}

		public double getDouble(Object graph) {
			Object o = get(graph);
			if (o == null)
				return 0.0;
			else
				return ((Double) o).doubleValue();
		}

		public int getInt(Object graph) {
			Object o = get(graph);
			if (o == null)
				return 0;
			else
				return ((Integer) o).intValue();
		}

		public void setBool(Object graph, boolean value) {
			set(graph, new Boolean(value));
		}

		public void setDouble(Object graph, double value) {
			set(graph, new Double(value));
		}

		public void setInt(Object graph, int value) {
			set(graph, new Integer(value));
		}
	}

	// GRAPH LISTENER ------------------------------------------------

	/**
	 * A utility class that creates unique IDs. It does so be keeping track of
	 * previously created IDs.
	 * 
	 * @author Werner Jainek
	 */
	protected class IDCreator {

		/**
		 * The lowest possible valid ID.
		 */
		private int startID;

		/**
		 * The highest ID that has been created up to this point.
		 */
		private int highestID;

		/**
		 * Instantiates a new <code>IDCreator</code> that assigns IDs in the
		 * range <code>[1,inf)</code>.
		 */
		public IDCreator() {
			startID = 1;
			highestID = 0;
		}

		/**
		 * Instantiates a new <code>IDCreator</code> that assigns IDs in the
		 * range <code>[startID,inf)</code>.
		 */
		public IDCreator(int startID) {
			this.startID = startID;
			this.highestID = startID - 1;
		}

		/**
		 * Returns a new and unique ID.
		 */
		public int createID() {
			++highestID;
			return highestID;
		}

		/**
		 * Marks the passed as ready for reuse.
		 */
		public void removeID(int id) {
			// TODO: add removed id to available list and reuse !
		}

		/**
		 * Tells the <code>IDCreator</code> that the passed ID should not be
		 * used for creating new IDs, since it is already used.
		 * 
		 * @param id
		 */
		public void setUsed(int id) {
			highestID = Math.max(highestID, id);
			// TODO: check available list
		}
	}

	/**
	 * A <code>GraphListener</code> that is responsible for assigning new and
	 * unique IDs to newly created nodes and edges in graphs that belong to this
	 * <code>LayeredGraph</code>.
	 * 
	 * @author Werner Jainek
	 */
	protected class IDAssigner implements GraphListener {

		/**
		 * This method is responsible for assigning new and unique IDs to nodes
		 * and edges whenever they are created and to remove IDs of nodes and
		 * edges whenever they are deleted from a graph.
		 * <p>
		 * Special care is taken if a node or edge is removed that has an ID
		 * that is still assigned to some other node or edge. In this case, the
		 * ID is not removed from the set of used IDs.
		 * <p>
		 * If a node ID is removed, all other nodes that have that ID as their
		 * parent ID are visited and their parent ID is set to <code>0</code>.
		 * 
		 * @see y.base.GraphListener#onGraphEvent(y.base.GraphEvent)
		 */
		public void onGraphEvent(GraphEvent evt) {
			Graph g = evt.getGraph();

			// CREATION
			if (evt.getType() == GraphEvent.NODE_CREATION) {
				Node n = (Node) evt.getData();

				y.base.NodeMap nID = (y.base.NodeMap) g
						.getDataProvider(NODE_ID_DPKEY);
				nID.setInt(n, nodeIDCreator.createID());

				y.base.NodeMap pID = (y.base.NodeMap) g
						.getDataProvider(PARENT_ID_DPKEY);
				pID.setInt(n, 0);
			}

			else if (evt.getType() == GraphEvent.EDGE_CREATION) {
				Edge e = (Edge) evt.getData();

				y.base.EdgeMap eID = (y.base.EdgeMap) g
						.getDataProvider(EDGE_ID_DPKEY);
				eID.setInt(e, edgeIDCreator.createID());
			}

			// REMOVAL -- TODO: this code could be wrong if nodes/edges are
			// hidden in the graphs
			else if (evt.getType() == GraphEvent.PRE_NODE_REMOVAL) {
				// we can't just remove the id, since another node might still
				// use it

				// get the ID of the removed Node
				Node removedNode = (Node) evt.getData();
				removeNodeID(removedNode);
			}

			else if (evt.getType() == GraphEvent.PRE_EDGE_REMOVAL) {
				// we can't just remove the id, since another edge might still
				// use it

				// get the ID of the removed Edge
				Edge removedEdge = (Edge) evt.getData();
				removeEdgeID(removedEdge);
			}

		}

		/**
		 * Removes IDs of nodes and edges that occur in the passed graph from
		 * the set of used IDs. The implementation of this method is very
		 * simple, since it only calls <code>removeNodeID</code> and
		 * <code>removeEdgeID</code> for every node and every edge in the
		 * passed graph. For details, consult the documentation of those
		 * methods.
		 * <p>
		 * It is irrelevant if the passed graph is still part of the
		 * LayeredGraph or if it has already been removed. It is only important
		 * that is still has all the structure information (node/parent/edge
		 * IDs).
		 * 
		 * @param graph
		 *            The graph that holds the nodes and edges whose IDs will be
		 *            removed.
		 */
		public void removeIDs(Graph graph) {
			// remove node ids
			for (NodeCursor nc = graph.nodes(); nc.ok(); nc.next()) {
				removeNodeID(nc.node());
			}

			// remove edge ids
			for (EdgeCursor ec = graph.edges(); ec.ok(); ec.next()) {
				removeEdgeID(ec.edge());
			}
		}

		/**
		 * Removes the edge ID of the passed edge from the set of used IDs. This
		 * will happen only if there is no other edge inside the LayeredGraph
		 * that has the same ID.
		 * <p>
		 * It is irrelevant if the passed edge is still part of the LayeredGraph
		 * or if it is already removed (together with its parent graph). It is
		 * only necessary that the edges parent graph still has structure
		 * information (node/parent/edge IDs).
		 * 
		 * @param edge
		 *            The edge whose ID will be removed.
		 */
		public void removeEdgeID(Edge edge) {
			Graph g = edge.getGraph();

			y.base.EdgeMap edgeIDMap = (y.base.EdgeMap) g
					.getDataProvider(EDGE_ID_DPKEY);
			int edgeID = edgeIDMap.getInt(edge);

			// first, check if another edge has the same ID
			for (GraphCursor gc = graphs(); gc.ok(); gc.next()) {
				y.base.EdgeMap eID = (y.base.EdgeMap) gc.graph()
						.getDataProvider(EDGE_ID_DPKEY);
				for (EdgeCursor ec = gc.graph().edges(); ec.ok(); ec.next()) {
					if (ec.edge() != edge && eID.getInt(ec.edge()) == edgeID) {
						// if there is a edge with the same ID, there is nothing
						// to do
						return;
					}
				}
			}

			// no other edge uses the same ID, so we tell the IDCreator that the
			// edge ID is now free for reuse
			edgeIDCreator.removeID(edgeID);
		}

		/**
		 * Removes the node ID of the passed node from the set of used IDs. This
		 * will happen only if there is no other node inside the LayeredGraph
		 * that has the same ID.
		 * <p>
		 * If the node ID is removed, all other nodes that have this ID as their
		 * parent ID will be visited and their parent ID will be set to
		 * <code>0</code>.
		 * <p>
		 * It is irrelevant if the passed node is still part of the LayeredGraph
		 * or if it is already removed (together with its parent graph). It is
		 * only necessary that the nodes parent graph still has structure
		 * information (node/parent/edge IDs).
		 * 
		 * @param node
		 *            The node whose ID should be removed.
		 */
		public void removeNodeID(Node node) {
			Graph g = node.getGraph();
			y.base.NodeMap nodeIDMap = (y.base.NodeMap) g
					.getDataProvider(NODE_ID_DPKEY);
			int nodeID = nodeIDMap.getInt(node);

			// first, check if another node has the same ID
			for (GraphCursor gc = graphs(); gc.ok(); gc.next()) {
				y.base.NodeMap nID = (y.base.NodeMap) gc.graph()
						.getDataProvider(NODE_ID_DPKEY);
				for (NodeCursor nc = gc.graph().nodes(); nc.ok(); nc.next()) {
					if (nc.node() != node && nID.getInt(nc.node()) == nodeID) {
						// if there is a node with the same ID, there is nothing
						// to do
						return;
					}
				}
			}

			// no other node has the ID, so we check if the node that has to be
			// removed is the parent of any other node
			for (GraphCursor gc = graphs(); gc.ok(); gc.next()) {
				y.base.NodeMap pID = (y.base.NodeMap) gc.graph()
						.getDataProvider(PARENT_ID_DPKEY);
				for (NodeCursor nc = gc.graph().nodes(); nc.ok(); nc.next()) {
					if (nc.node() != node && pID.getInt(nc.node()) == nodeID) {
						// if there is a node with the node to be removed as a
						// parent, we remove the parent
						pID.setInt(nc.node(), 0);
					}
				}
			}

			// finally, we tell the IDCreator that the node ID is now free for
			// reuse
			nodeIDCreator.removeID(nodeID);
		}

	}

	// DEBUGGING ------------------------------------------------
	/**
	 * Prints a string representation of this <code>LayeredGraph</code>.
	 * 
	 * @see java.lang.Object#toString()
	 */
	public String toString() {

		String s = "";

		int i = 0;
		for (GraphCursor gc = graphs(); gc.ok(); gc.next()) {
			s += "Layer " + i + ":\n";

			s += "Node|ParentIDs: ";
			DataProvider nID = gc.graph().getDataProvider(NODE_ID_DPKEY);
			DataProvider pID = gc.graph().getDataProvider(PARENT_ID_DPKEY);
			for (NodeCursor nc = gc.graph().nodes(); nc.ok(); nc.next()) {
				s += "(" + nID.getInt(nc.node()) + "|" + pID.getInt(nc.node())
						+ ")";
				s += " ";
			}
			s += "\n";

			s += "EdgeIDs: ";
			DataProvider eID = gc.graph().getDataProvider(EDGE_ID_DPKEY);
			for (EdgeCursor ec = gc.graph().edges(); ec.ok(); ec.next()) {
				s += "(" + eID.getInt(ec.edge()) + ")";
				s += " ";
			}
			s += "\n";

			s += "\n";
			++i;
		}

		// LAyer edges
		HashMap graphOrder = new HashMap();
		i = 0;
		for (GraphCursor gc = graphs(); gc.ok(); gc.next()) {
			graphOrder.put(gc.graph(), new Integer(i));
			++i;
		}

		s += "\n";
		i = 0;
		for (LayerEdgeCursor lec = layerEdges(); lec.ok(); lec.next()) {
			LayerEdge le = lec.layerEdge();

			Integer srcGraph = (Integer) graphOrder.get(le.getSourceGraph());
			Integer trgGraph = (Integer) graphOrder.get(le.getTargetGraph());

			s += "LayerEdge " + i + ": ";
			s += "(Layer " + srcGraph.intValue() + ", Node "
					+ le.source().index() + ")";
			s += " -> ";
			s += "(Layer " + trgGraph.intValue() + ", Node "
					+ le.target().index() + ")";
			s += "\n";

			++i;
		}

		return s;
	}
}
