/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.graphics;

/**
 * Represents a transform in three dimensional space that can be applied to
 * points and vectors.
 * <p>
 * Internally, every transform is stored as a 4x4 matrix <code>M</code>.
 * Applying a transform on a <em>point</em> <code>p = (px,py,pz)^t</code>
 * simply means pre-multiplying the matrix after extending the point with the
 * homogeneous coordinate of <code>1</code>:
 * <code>p_transformed = M * (px,py,pz,1)^t</code> ("^t" means transposed).
 * <p>
 * Similary, transforming a <em>vector</em> <code>v = (vx,vy,vz)</code> is
 * accomplished by appending a homogeneous coordinate of <code>0</code> and
 * then pre-multiplying by <code>M</code>:
 * <code>v_transformed = M * (vx,vy,vz,0)^t</code>.
 * <p>
 * Therefore, all transformations that can be performed by this class are affine
 * and projective transforms.
 * 
 * @author Werner Jainek
 */
public class Transform {

	/**
	 * Creates a transform, that, when applied, performs a translation by the
	 * given vector.
	 */
	public static Transform translate(Vector3D delta) {
		Matrix4x4 m = new Matrix4x4(1, 0, 0, delta.x, 0, 1, 0, delta.y, 0, 0,
				1, delta.z, 0, 0, 0, 1);
		Matrix4x4 minv = new Matrix4x4(1, 0, 0, -delta.x, 0, 1, 0, -delta.y, 0,
				0, 1, -delta.z, 0, 0, 0, 1);
		return new Transform(m, minv);
	}

	/**
	 * Creates a transform, that, when applied, performs a scaling of the
	 * x,y,z-axes by the given factors.
	 */
	public static Transform scale(float x, float y, float z) {
		Matrix4x4 m = new Matrix4x4(x, 0, 0, 0, 0, y, 0, 0, 0, 0, z, 0, 0, 0,
				0, 1);
		Matrix4x4 minv = new Matrix4x4(1.f / x, 0, 0, 0, 0, 1.f / y, 0, 0, 0,
				0, 1.f / z, 0, 0, 0, 0, 1);
		return new Transform(m, minv);
	}

	/**
	 * Creates a transform, that, when applied, performs a counter-clockwise
	 * rotation around the x-axis. More precisely, when looking down the
	 * negative x-axis, the rotation that is performed is counter-clockwise.
	 * <p>
	 * The passed angles are in degrees, not radians.
	 */
	public static Transform rotateX(float angle) {
		float rad = (float) Math.toRadians(angle);
		float s = (float) Math.sin(rad);
		float c = (float) Math.cos(rad);
		Matrix4x4 rot = new Matrix4x4(1, 0, 0, 0, 0, c, -s, 0, 0, s, c, 0, 0,
				0, 0, 1);
		return new Transform(rot, rot.getTransposed());
	}

	/**
	 * Creates a transform, that, when applied, performs a counter-clockwise
	 * rotation around the y-axis. More precisely, when looking down the
	 * negative y-axis, the rotation that is performed is counter-clockwise.
	 * <p>
	 * The passed angles are in degrees, not radians.
	 */
	public static Transform rotateY(float angle) {
		float rad = (float) Math.toRadians(angle);
		float s = (float) Math.sin(rad);
		float c = (float) Math.cos(rad);
		Matrix4x4 rot = new Matrix4x4(c, 0, s, 0, 0, 1, 0, 0, -s, 0, c, 0, 0,
				0, 0, 1);
		return new Transform(rot, rot.getTransposed());

	}

	/**
	 * Creates a transform, that, when applied, performs a counter-clockwise
	 * rotation around the z-axis. More precisely, when looking down the
	 * negative z-axis, the rotation that is performed is counter-clockwise.
	 * <p>
	 * The passed angles are in degrees, not radians.
	 */
	public static Transform rotateZ(float angle) {
		float rad = (float) Math.toRadians(angle);
		float s = (float) Math.sin(rad);
		float c = (float) Math.cos(rad);
		Matrix4x4 rot = new Matrix4x4(c, -s, 0, 0, s, c, 0, 0, 0, 0, 1, 0, 0,
				0, 0, 1);
		return new Transform(rot, rot.getTransposed());
	}

	/**
	 * Creates a transform, that, when applied, performs a counter-clockwise
	 * rotation around the given vector. More precisely, when looking in the
	 * opposite direction as the passed vector, the rotation performed is
	 * counter-clockwise.
	 * <p>
	 * The passed angles are in degrees, not radians.
	 */
	public static Transform rotate(Vector3D axis, float angle) {
		Vector3D a = axis.getNormalized();
		float rad = (float) Math.toRadians(angle);
		float s = (float) Math.sin(rad);
		float c = (float) Math.cos(rad);

		Matrix4x4 mat = new Matrix4x4();

		mat.m[0][0] = a.x * a.x + (1.f - a.x * a.x) * c;
		mat.m[0][1] = a.x * a.y * (1.f - c) - a.z * s;
		mat.m[0][2] = a.x * a.z * (1.f - c) + a.y * s;
		mat.m[0][3] = 0;

		mat.m[1][0] = a.x * a.y * (1.f - c) + a.z * s;
		mat.m[1][1] = a.y * a.y + (1.f - a.y * a.y) * c;
		mat.m[1][2] = a.y * a.z * (1.f - c) - a.x * s;
		mat.m[1][3] = 0;

		mat.m[2][0] = a.x * a.z * (1.f - c) - a.y * s;
		mat.m[2][1] = a.y * a.z * (1.f - c) + a.x * s;
		mat.m[2][2] = a.z * a.z + (1.f - a.z * a.z) * c;
		mat.m[2][3] = 0;

		mat.m[3][0] = 0;
		mat.m[3][1] = 0;
		mat.m[3][2] = 0;
		mat.m[3][3] = 1;

		return new Transform(mat, mat.getTransposed());
	}

	/**
	 * Creates a transform that can be used as a viewing transform that has the
	 * same effect as the OpenGL gluLookAt() routine.
	 * <p>
	 * When a "viewing transform" is applied to a vector that has world
	 * coordinates, they are transformed to camera coordinates.
	 */
	public static Transform lookAt(Point3D pos, Point3D look, Vector3D up) {
		Matrix4x4 camToWorld = new Matrix4x4();
		// Initialize fourth column of viewing matrix
		camToWorld.m[0][3] = pos.x;
		camToWorld.m[1][3] = pos.y;
		camToWorld.m[2][3] = pos.z;
		camToWorld.m[3][3] = 1;

		// Initialize first three columns of viewing matrix
		Vector3D dir = new Vector3D(pos, look).normalize();
		Vector3D right = dir.getCrossed(up).normalize();
		Vector3D newUp = right.getCrossed(dir);
		camToWorld.m[0][0] = right.x;
		camToWorld.m[1][0] = right.y;
		camToWorld.m[2][0] = right.z;
		camToWorld.m[3][0] = 0.0f;
		camToWorld.m[0][1] = newUp.x;
		camToWorld.m[1][1] = newUp.y;
		camToWorld.m[2][1] = newUp.z;
		camToWorld.m[3][1] = 0.0f;
		camToWorld.m[0][2] = dir.x;
		camToWorld.m[1][2] = dir.y;
		camToWorld.m[2][2] = dir.z;
		camToWorld.m[3][2] = 0.0f;

		return new Transform(camToWorld.getInverted(), camToWorld);
	}

	/**
	 * Creates a transform that can be used as a projection transform that has
	 * the same effect as the OpenGL gluPerspective() routine.
	 */
	public static Transform perspective(float fov, float aspect, float zNear,
			float zFar) {
		if (zNear == zFar) {
			zFar += 0.1f;
		}
		if (fov == 180.0f) {
			fov = 179.0f;
		}
		if (aspect == 0) {
			aspect = 1.0f;
		}
		// // Perform projective divide
		// float inv_denom = 1.0f / (zFar - zNear);
		// Matrix4x4 persp =
		// new Matrix4x4(1, 0, 0, 0,
		// 0, 1, 0, 0,
		// 0, 0, zFar*inv_denom, -zFar*zNear*inv_denom,
		// 0, 0, 1, 0);
		// // Scale to canonical viewing volume
		// float invTanAng = 1.0f / (float)Math.tan(Math.toRadians(fov) / 2.0f);
		// return scale(invTanAng/aspect, invTanAng, 1.0f).apply(
		// new Transform(persp) );

		// matrix copied from gluPerspective man page
		float n_f = zNear - zFar;
		float invTanAng = 1.0f / (float) Math.tan(Math.toRadians(fov) / 2.0f);
		Matrix4x4 persp = new Matrix4x4(invTanAng / aspect, 0, 0, 0, 0,
				invTanAng, 0, 0, 0, 0, (zNear + zFar) / n_f, 2 * zFar * zNear
						/ n_f, 0, 0, -1, 0);

		return new Transform(persp);
	}

	/**
	 * Creates a transform that can be used as a viewport transform that has the
	 * same effect as the OpenGL glViewport() routine. It leaves the
	 * z-coordinates intact and only remaps x/y coordinates.
	 */
	public static Transform viewport(float x, float y, float width, float height) {
		Matrix4x4 persp = new Matrix4x4(width / 2, 0, 0, width / 2 * (x + 1),
				0, height / 2, 0, height / 2 * (y + 1), 0, 0, 1, 0, 0, 0, 0, 1);

		return new Transform(persp);
	}

	/**
	 * The matrix that stores the current transform operation.
	 */
	protected Matrix4x4 transform;

	/**
	 * The matrix that stores the inverse of the current transform operation.
	 */
	protected Matrix4x4 inverse;

	/**
	 * Storage specifier - indicates that the rows of a matrix are stored
	 * consecutively.
	 */
	public static final int ROWS = 0;

	/**
	 * Storage specifier - indicates that the columns of a matrix are stored
	 * consecutively.
	 */
	public static final int COLUMNS = 1;

	/**
	 * Creates an identity transform (a transform that has no effect when
	 * applied).
	 */
	public Transform() {
		transform = new Matrix4x4();
		transform.m[0][0] = 1;
		transform.m[1][1] = 1;
		transform.m[2][2] = 1;
		transform.m[3][3] = 1;
		inverse = transform;
	}

	/**
	 * Creates a transform by copying the given transform.
	 */
	public Transform(Transform t) {
		transform = t.transform;
		inverse = t.inverse;
	}

	/**
	 * Creates a transform by means of a user specified transformation matrix.
	 * <p>
	 * The passed array has to hold 4*4=16 values for all entries of a 4x4
	 * matrix and the <code>storeOrder</code> specifies if columns or rows
	 * appear consecutively in the passed array.
	 */
	public Transform(float[] m, int storeOrder) {
		transform = new Matrix4x4(m);
		if (storeOrder == COLUMNS) {
			transform.transpose();
		}
		inverse = transform.getInverted();
	}

	/**
	 * Creates a transform by means of a user specified transformation matrix.
	 */
	public Transform(Matrix4x4 trans) {
		transform = new Matrix4x4(trans);
		inverse = transform.getInverted();
	}

	/**
	 * Creates a transform by means of a user specified transformation matrix
	 * and its inverse.
	 */
	public Transform(Matrix4x4 trans, Matrix4x4 inv) {
		transform = trans;
		inverse = inv;
	}

	/**
	 * Sets the current transform to the identity transform.
	 */
	public void setIdentity() {
		transform = new Matrix4x4();
		transform.m[0][0] = 1;
		transform.m[1][1] = 1;
		transform.m[2][2] = 1;
		transform.m[3][3] = 1;
		inverse = transform;
	}

	/**
	 * Returns the entries of the matrix that represents this transform in the
	 * specified order.
	 */
	public float[] getArray(int storeOrder) {
		float a[] = new float[16];

		switch (storeOrder) {

		case ROWS: {
			int k = 0;
			for (int i = 0; i < 4; ++i) {
				for (int j = 0; j < 4; ++j) {
					a[k] = transform.m[i][j];
					++k;
				}
			}
		}
			break;

		default:
		case COLUMNS: {
			int k = 0;
			for (int i = 0; i < 4; ++i) {
				for (int j = 0; j < 4; ++j) {
					a[k] = transform.m[j][i];
					++k;
				}
			}
		}
			break;
		}
		return a;
	}

	/**
	 * Inverts this transform. Returns a reference to the inverted transform.
	 */
	public Transform invert() {
		Matrix4x4 tmp = transform;
		transform = inverse;
		inverse = tmp;
		return this;
	}

	/**
	 * Returns copy of this transform that is the inverse of this transform.
	 */
	public Transform getInverted() {
		return new Transform(inverse, transform);
	}

	/**
	 * Applies this transform to the passed point and returns a reference to
	 * that point.
	 */
	public Point3D apply(Point3D p) {
		float x = p.x, y = p.y, z = p.z;
		p.x = transform.m[0][0] * x + transform.m[0][1] * y + transform.m[0][2]
				* z + transform.m[0][3];
		p.y = transform.m[1][0] * x + transform.m[1][1] * y + transform.m[1][2]
				* z + transform.m[1][3];
		p.z = transform.m[2][0] * x + transform.m[2][1] * y + transform.m[2][2]
				* z + transform.m[2][3];
		float pw = transform.m[3][0] * x + transform.m[3][1] * y
				+ transform.m[3][2] * z + transform.m[3][3];

		if (pw == 0) {
			System.err
					.println("Error: Used a transformation in Transform.apply(Point3D) that would send the point to infinity. Keeping old point.\n");
			p.x = x;
			p.y = y;
			p.z = z;
			return p;
		}

		else if (pw == 1.) {
			return p;
		}

		else {
			p.x /= pw;
			p.y /= pw;
			p.z /= pw;
			return p;
		}
	}

	/**
	 * Applies this transform to a copy of the passed point and returns the
	 * copy.
	 */
	public Point3D getApplied(Point3D p) {
		Point3D result = new Point3D(p);
		return this.apply(result);
	}

	/**
	 * Applies this transform to the passed vector and returns a reference to
	 * that vector.
	 */
	public Vector3D apply(Vector3D v) {
		float x = v.x, y = v.y, z = v.z;
		v.x = transform.m[0][0] * x + transform.m[0][1] * y + transform.m[0][2]
				* z + transform.m[0][3];
		v.y = transform.m[1][0] * x + transform.m[1][1] * y + transform.m[1][2]
				* z + transform.m[1][3];
		v.z = transform.m[2][0] * x + transform.m[2][1] * y + transform.m[2][2]
				* z + transform.m[2][3];

		if (transform.m[3][0] + transform.m[3][1] + transform.m[3][2] > 0.0f) {
			System.err
					.println("Error: Used a transformation in Transform.apply(Vector3D) that would project the direction(vector) to a point in 3d-space. Keeping old direction.\n");
			v.x = x;
			v.y = y;
			v.z = z;
			return v;
		}

		return v;
	}

	/**
	 * Applies this transform to the passed vector and returns a reference to
	 * that vector.
	 */
	public Vector3D getApplied(Vector3D v) {
		Vector3D result = new Vector3D(v);
		return this.apply(result);
	}

	/**
	 * Applies this transform to the passed transform and returns a reference to
	 * the modified passed transform.
	 */
	public Transform apply(Transform t) {
		Matrix4x4 newTrans = new Matrix4x4(t.transform);
		transform.apply(newTrans);

		Matrix4x4 newInv = new Matrix4x4(inverse);
		t.inverse.apply(newInv);

		t.transform = newTrans;
		t.inverse = newInv;

		return t;
	}

	/**
	 * Applies this transform to a copy of the passed transform and returns a
	 * reference to that copy.
	 */
	public Transform getApplied(Transform t) {
		Transform res = new Transform(t);
		return this.apply(res);
	}

	public String toString() {
		return "Transform\n" + transform.toString() + "Inverse\n"
				+ inverse.toString();
	}

	/**
	 * A representation for 4x4 matrices.
	 * 
	 * @author Werner Jainek
	 */
	public static class Matrix4x4 {

		public float m[][];

		public Matrix4x4() {
			m = new float[4][4];
		}

		public Matrix4x4(float t00, float t01, float t02, float t03, float t10,
				float t11, float t12, float t13, float t20, float t21,
				float t22, float t23, float t30, float t31, float t32, float t33) {
			m = new float[4][4];
			m[0][0] = t00;
			m[0][1] = t01;
			m[0][2] = t02;
			m[0][3] = t03;
			m[1][0] = t10;
			m[1][1] = t11;
			m[1][2] = t12;
			m[1][3] = t13;
			m[2][0] = t20;
			m[2][1] = t21;
			m[2][2] = t22;
			m[2][3] = t23;
			m[3][0] = t30;
			m[3][1] = t31;
			m[3][2] = t32;
			m[3][3] = t33;
		}

		public Matrix4x4(float[] mat) {
			m = new float[4][4];
			if (mat.length == 16) {
				int k = 0;
				for (int i = 0; i < 4; ++i) {
					for (int j = 0; j < 4; ++j) {
						m[i][j] = mat[k];
						++k;
					}
				}
			}
		}

		public Matrix4x4(Matrix4x4 mat) {
			m = new float[4][4];
			for (int i = 0; i < 4; ++i) {
				for (int j = 0; j < 4; ++j) {
					m[i][j] = mat.m[i][j];
				}
			}
		}

		public String toString() {
			return "[ " + m[0][0] + "\t" + m[0][1] + "\t" + m[0][2] + "\t"
					+ m[0][3] + "\n" + "  " + m[1][0] + "\t" + m[1][1] + "\t"
					+ m[1][2] + "\t" + m[1][3] + "\n" + "  " + m[2][0] + "\t"
					+ m[2][1] + "\t" + m[2][2] + "\t" + m[2][3] + "\n" + "  "
					+ m[3][0] + "\t" + m[3][1] + "\t" + m[3][2] + "\t"
					+ m[3][3] + " ]\n";
		}

		public Matrix4x4 transpose() {
			float tm01 = m[0][1], tm02 = m[0][2], tm03 = m[0][3];
			float tm12 = m[1][2], tm13 = m[1][3];
			float tm23 = m[2][3];

			m[0][1] = m[1][0];
			m[1][0] = tm01;
			m[0][2] = m[2][0];
			m[2][0] = tm02;
			m[0][3] = m[3][0];
			m[3][0] = tm03;

			m[1][2] = m[2][1];
			m[2][1] = tm12;
			m[1][3] = m[3][1];
			m[3][1] = tm13;

			m[2][3] = m[3][2];
			m[3][2] = tm23;

			return this;
		}

		public Matrix4x4 getTransposed() {
			return new Matrix4x4(m[0][0], m[1][0], m[2][0], m[3][0], m[0][1],
					m[1][1], m[2][1], m[3][1], m[0][2], m[1][2], m[2][2],
					m[3][2], m[0][3], m[1][3], m[2][3], m[3][3]);
		}

		public Matrix4x4 getInverted() {
			Matrix4x4 i = new Matrix4x4(this);
			return i.invert();
		}

		public Matrix4x4 invert() { // from pbrt.org
			int indxc[] = new int[4];
			int indxr[] = new int[4];
			int ipiv[] = new int[4];
			float minv[][] = new float[4][4];

			for (int i = 0; i < 4; ++i)
				for (int j = 0; j < 4; ++j)
					minv[i][j] = m[i][j];

			for (int i = 0; i < 4; i++) {
				int irow = -1, icol = -1;
				float big = 0.0f;
				// Choose pivot
				for (int j = 0; j < 4; j++) {
					if (ipiv[j] != 1) {
						for (int k = 0; k < 4; k++) {
							if (ipiv[k] == 0) {
								if (Math.abs(minv[j][k]) >= big) {
									big = Math.abs(minv[j][k]);
									irow = j;
									icol = k;
								}
							} else if (ipiv[k] > 1) {
								System.err
										.println("Error: Singular matrix in Matrix4x4.inverse");
								return this;
							}
						}
					}
				}
				++ipiv[icol];
				// Swap rows irow and icol for pivot
				if (irow != icol) {
					for (int k = 0; k < 4; ++k) {
						float tmp = minv[irow][k];
						minv[irow][k] = minv[icol][k];
						minv[icol][k] = tmp;
					}
				}
				indxr[i] = irow;
				indxc[i] = icol;
				if (minv[icol][icol] == 0.)
					System.err
							.println("Error: Singular matrix in Matrix4x4.inverse");
				// Set m[icol][icol] to one by scaling row _icol_ appropriately
				float pivinv = 1.f / minv[icol][icol];
				minv[icol][icol] = 1.f;
				for (int j = 0; j < 4; j++)
					minv[icol][j] *= pivinv;
				// Subtract this row from others to zero out their columns
				for (int j = 0; j < 4; j++) {
					if (j != icol) {
						float save = minv[j][icol];
						minv[j][icol] = 0;
						for (int k = 0; k < 4; k++)
							minv[j][k] -= minv[icol][k] * save;
					}
				}
			}
			// Swap columns to reflect permutation
			for (int j = 3; j >= 0; j--) {
				if (indxr[j] != indxc[j]) {
					for (int k = 0; k < 4; k++) {
						float tmp = minv[k][indxr[j]];
						minv[k][indxr[j]] = minv[k][indxc[j]];
						minv[k][indxc[j]] = tmp;
					}
				}
			}
			m = minv;
			return this;
		}

		public Matrix4x4 getApplied(Matrix4x4 mat) {
			Matrix4x4 result = new Matrix4x4(mat);
			return this.apply(result);
		}

		public Matrix4x4 apply(Matrix4x4 mat) {
			float t0, t1, t2, t3;

			t0 = m[0][0] * mat.m[0][0] + m[0][1] * mat.m[1][0] + m[0][2]
					* mat.m[2][0] + m[0][3] * mat.m[3][0];
			t1 = m[1][0] * mat.m[0][0] + m[1][1] * mat.m[1][0] + m[1][2]
					* mat.m[2][0] + m[1][3] * mat.m[3][0];
			t2 = m[2][0] * mat.m[0][0] + m[2][1] * mat.m[1][0] + m[2][2]
					* mat.m[2][0] + m[2][3] * mat.m[3][0];
			t3 = m[3][0] * mat.m[0][0] + m[3][1] * mat.m[1][0] + m[3][2]
					* mat.m[2][0] + m[3][3] * mat.m[3][0];
			mat.m[0][0] = t0;
			mat.m[1][0] = t1;
			mat.m[2][0] = t2;
			mat.m[3][0] = t3;

			t0 = m[0][0] * mat.m[0][1] + m[0][1] * mat.m[1][1] + m[0][2]
					* mat.m[2][1] + m[0][3] * mat.m[3][1];
			t1 = m[1][0] * mat.m[0][1] + m[1][1] * mat.m[1][1] + m[1][2]
					* mat.m[2][1] + m[1][3] * mat.m[3][1];
			t2 = m[2][0] * mat.m[0][1] + m[2][1] * mat.m[1][1] + m[2][2]
					* mat.m[2][1] + m[2][3] * mat.m[3][1];
			t3 = m[3][0] * mat.m[0][1] + m[3][1] * mat.m[1][1] + m[3][2]
					* mat.m[2][1] + m[3][3] * mat.m[3][1];
			mat.m[0][1] = t0;
			mat.m[1][1] = t1;
			mat.m[2][1] = t2;
			mat.m[3][1] = t3;

			t0 = m[0][0] * mat.m[0][2] + m[0][1] * mat.m[1][2] + m[0][2]
					* mat.m[2][2] + m[0][3] * mat.m[3][2];
			t1 = m[1][0] * mat.m[0][2] + m[1][1] * mat.m[1][2] + m[1][2]
					* mat.m[2][2] + m[1][3] * mat.m[3][2];
			t2 = m[2][0] * mat.m[0][2] + m[2][1] * mat.m[1][2] + m[2][2]
					* mat.m[2][2] + m[2][3] * mat.m[3][2];
			t3 = m[3][0] * mat.m[0][2] + m[3][1] * mat.m[1][2] + m[3][2]
					* mat.m[2][2] + m[3][3] * mat.m[3][2];
			mat.m[0][2] = t0;
			mat.m[1][2] = t1;
			mat.m[2][2] = t2;
			mat.m[3][2] = t3;

			t0 = m[0][0] * mat.m[0][3] + m[0][1] * mat.m[1][3] + m[0][2]
					* mat.m[2][3] + m[0][3] * mat.m[3][3];
			t1 = m[1][0] * mat.m[0][3] + m[1][1] * mat.m[1][3] + m[1][2]
					* mat.m[2][3] + m[1][3] * mat.m[3][3];
			t2 = m[2][0] * mat.m[0][3] + m[2][1] * mat.m[1][3] + m[2][2]
					* mat.m[2][3] + m[2][3] * mat.m[3][3];
			t3 = m[3][0] * mat.m[0][3] + m[3][1] * mat.m[1][3] + m[3][2]
					* mat.m[2][3] + m[3][3] * mat.m[3][3];
			mat.m[0][3] = t0;
			mat.m[1][3] = t1;
			mat.m[2][3] = t2;
			mat.m[3][3] = t3;

			return mat;
		}

	}

}
