/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.layout;

import java.util.Collections;
import java.util.List;
import java.util.Vector;

/**
 * A layouter that allows to express a layout operation as a chain of multiple
 * layout stages.
 * 
 * 
 * This interface mirrors the <code>CompositeLayouter</code> interface found
 * in the <code>y.layout</code> package.
 * 
 * @author Werner Jainek
 */
public class CompositeLayouter25D implements Layouter25D {

	/**
	 * The layouter that is always invoked at the end of the chain.
	 */
	protected Layouter25D coreLayouter;

	/**
	 * The first layout stage.
	 */
	protected LayoutStage25D firstStage;

	/**
	 * The last layout stage that is invoked right before the core layouter is
	 * invoked.
	 */
	protected LayoutStage25D lastStage;

	/**
	 * Creates a new instance of <code>CompositeLayouter25D</code>.
	 * 
	 * @param stage
	 *            the layout stage that is wrapped around the core layouter
	 * @param coreLayouter
	 *            the core layouter of this composite layouter
	 */
	public CompositeLayouter25D(LayoutStage25D stage, Layouter25D coreLayouter) {
		this.coreLayouter = coreLayouter;
		firstStage = stage;
		lastStage = stage;
		stage.setCoreLayouter25D(coreLayouter);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see y25.layout.Layouter25D#canLayout(y25.layout.LayeredLayoutGraph)
	 */
	public boolean canLayout(LayeredLayoutGraph lg) {
		return firstStage.canLayout(lg);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see y25.layout.Layouter25D#doLayout(y25.layout.LayeredLayoutGraph)
	 */
	public void doLayout(LayeredLayoutGraph lg) {
		firstStage.doLayout(lg);
	}

	/**
	 * Appends a stage to the layout pipeline. Stages added with this method
	 * will be invoked just before the core layouter of the composite layouter
	 * will be invoked.
	 */
	public void appendStage(LayoutStage25D stage) {
		stage.setCoreLayouter25D(firstStage);
		firstStage = stage;
	}

	/**
	 * Prepends a stage to the layout pipeline. Stages added with this method
	 * will be invoked before any other stages will be invoked.
	 */
	public void prependStage(LayoutStage25D stage) {
		stage.setCoreLayouter25D(coreLayouter);
		lastStage.setCoreLayouter25D(stage);
		lastStage = stage;
	}

	/**
	 * Returns a vector of all layout stages already present in the layout
	 * chain. The first entry of the vector is the first layout stage, the last
	 * entry is the last layout stage that is invoked before the core layouter
	 * is invoked.
	 */
	public List getLayoutStages() {
		Vector v = new Vector();

		// gather all layout stages
		Layouter25D current = firstStage;

		while (current != coreLayouter) {
			v.add(current);
			current = ((LayoutStage25D) current).getCoreLayouter25D();
		}

		// return the list
		Collections.reverse(v);
		return v;
	}

}
