/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.view;

import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;
import java.nio.FloatBuffer;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;

import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLCanvas;
import javax.media.opengl.GLEventListener;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.Graph;
import y.base.NodeCursor;
import y.view.Graph2D;
import y25.base.GraphCursor;
import y25.base.LayerEdgeCursor;
import y25.graphics.Camera;

import com.sun.opengl.util.Animator;

/**
 * Displays and navigates a <code>Graph25D</code>.
 * <p>
 * Usually, this class is used in the following way:
 * 
 * <pre>
 * // 1. create a Graph25D
 * Graph25D graph25d = new Graph25D();
 * 
 * // 2. create a Graph25DView and associate view modes with it
 * Graph25DView view25d = new Graph25DView();
 * view25d.setGraph25D(graph25d);
 * view25d.addViewMode25D(new NavigationMode25D());
 * 
 * // 3. create a panel and put the view inside
 * JPanel panel = new JPanel();
 * panel.add(view25d.getGLCanvas());
 * </pre>
 * 
 * @author Werner Jainek
 */
public class Graph25DView implements GLEventListener {

	/**
	 * The OpenGL canvas that this view uses for rendering.
	 */
	protected GLCanvas glCanvas;

	/**
	 * The animator that drives the redraw of the entire scene every frame.
	 */
	protected Animator animator;

	/**
	 * Handles the viewing transform.
	 */
	protected Camera camera;

	/**
	 * Stores the view modes associated with this view.
	 */
	protected LinkedList viewModes;

	/**
	 * A reference to the graph that is displayed.
	 */
	protected Graph25D graph25D;

	/**
	 * Temporary storage for screenshot data.
	 */
	private FloatBuffer picturePixelData;

	/**
	 * Tells {@link #display display()} whether or not to take a screenshot the
	 * next time a frame is rendered.
	 */
	private boolean takePicture = false; // signals if next rendered frame

	// should be copied to the above
	// FloatBuffer

	/**
	 * Creates an empty <code>Graph25DView</code>.
	 */
	public Graph25DView() {

		// Create Canvas
		// old JOGL: glCanvas = GLDrawableFactory.getFactory().createGLCanvas(
		// new GLCapabilities() );
		glCanvas = new GLCanvas();

		// Attach GLEventListener and create an Animator
		glCanvas.addGLEventListener(this);
		animator = new Animator(glCanvas);
		animator.start();

		// Create Camera
		camera = new Camera();

		// Create ViewModes
		viewModes = new LinkedList();
	}

	/**
	 * Returns the OpenGL canvas this view renders into.
	 */
	public GLCanvas getGLCanvas() {
		return glCanvas;
	}

	/**
	 * Returns the camera this view uses for displaying the scene.
	 */
	public Camera getCamera() {
		return camera;
	}

	/**
	 * Sets the graph that is displayed in this view.
	 */
	public void setGraph25D(Graph25D g) {
		graph25D = g;
	}

	/**
	 * Returns the graph that is displayed in this view. Returns
	 * <code>null</code> if no graph is associated with the view.
	 */
	public Graph25D getGraph25D() {
		return graph25D;
	}

	/**
	 * Takes a screenshot of the current view and returns it.
	 */
	public BufferedImage getPicture() {

		// 1st Display the scene and record all pixel values
		takePicture = true;
		glCanvas.display();

		// 2nd Create an image out of pixel data
		int width = glCanvas.getWidth();
		int height = glCanvas.getHeight();
		BufferedImage img = new BufferedImage(width, height,
				BufferedImage.TYPE_INT_RGB);
		WritableRaster r = ((WritableRaster) img.getRaster());

		int tmp[] = new int[3];
		int idx = 0;
		for (int j = 0; j < height; ++j) {
			for (int i = 0; i < width; ++i) {
				tmp[0] = (int) (255 * picturePixelData.get(idx * 3 + 0));
				tmp[1] = (int) (255 * picturePixelData.get(idx * 3 + 1));
				tmp[2] = (int) (255 * picturePixelData.get(idx * 3 + 2));
				r.setPixel(i, height - j - 1, tmp);
				++idx;
			}
		}
		return img;
	}

	// --- GLEventListener Methods ----

	/**
	 * Initializes the view.
	 * <p>
	 * Sets the standard lights, whether or not to use antialiasing and various
	 * other OpenGL states.
	 * <p>
	 * This method does not have to be called directly, since that is handled by
	 * the Java OpenGL system.
	 * 
	 * @see javax.media.opengl.GLEventListener#init(javax.media.opengl.GLAutoDrawable)
	 */
	public void init(GLAutoDrawable drawable) {
		GL gl = drawable.getGL();

		// State Machine Settings
		gl.setSwapInterval(1);

		// -- display quality
		// gl.glEnable( GL.GL_MULTISAMPLE ); // not supported yet
		gl.glEnable(GL.GL_LINE_SMOOTH);

		// -- geometry
		gl.glEnable(GL.GL_CULL_FACE);
		gl.glEnable(GL.GL_DEPTH_TEST);
		gl.glEnable(GL.GL_NORMALIZE);

		// -- alpha blending
		gl.glEnable(GL.GL_BLEND);
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
		gl.glAlphaFunc(GL.GL_GREATER, 0.1f);

		// Projection Settings
		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glLoadIdentity();

		gl.glMatrixMode(GL.GL_MODELVIEW);
		gl.glLoadIdentity();

		// Light
		gl.glEnable(GL.GL_LIGHTING);
		float lmodel_ambient[] = { 0.7f, 0.7f, 0.7f, 1.0f };
		gl.glLightModelfv(GL.GL_LIGHT_MODEL_AMBIENT, lmodel_ambient, 0);

		gl.glEnable(GL.GL_LIGHT0);
		float light0_ambient[] = { 0.0f, 0.0f, 0.0f, 1.0f };
		float light0_diffuse[] = { 0.2f, 0.2f, 0.2f, 1.0f };
		float light0_specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_AMBIENT, light0_ambient, 0);
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_DIFFUSE, light0_diffuse, 0);
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_SPECULAR, light0_specular, 0);

		gl.glEnable(GL.GL_LIGHT1);
		float light1_ambient[] = { 0.0f, 0.0f, 0.0f, 1.0f };
		float light1_diffuse[] = { 0.2f, 0.2f, 0.2f, 1.0f };
		float light1_specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
		gl.glLightfv(GL.GL_LIGHT1, GL.GL_AMBIENT, light1_ambient, 0);
		gl.glLightfv(GL.GL_LIGHT1, GL.GL_DIFFUSE, light1_diffuse, 0);
		gl.glLightfv(GL.GL_LIGHT1, GL.GL_SPECULAR, light1_specular, 0);

		// Materials
		gl.glEnable(GL.GL_COLOR_MATERIAL);
		gl.glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

		// "camera" light
		// float pos[] = { 0.0f, 0.0f, 10.0f, 0.0f };
		// gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, pos);

	}

	/**
	 * Renders the scene into the given drawable.
	 * <p>
	 * This method does not have to be called directly, since that is handled by
	 * the Java OpenGL system.
	 * 
	 * @see javax.media.opengl.GLEventListener#display(javax.media.opengl.GLAutoDrawable)
	 */
	public void display(GLAutoDrawable drawable) {
		GL gl = drawable.getGL();

		// Clear the Screen
		gl.glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
		gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);

		// Adjust the Clipping Planes
		if (graph25D != null) {
			// TODO: remove this bb reclalculation and do it smarter (altough
			// this would probably be only a small speed improvement)
			graph25D.recalculateBoundingBox();
			camera.setClippingPlanes(graph25D.getBoundingBox());
		}

		// Activate the Projection
		camera.activateProjection(drawable);

		// Turn on the light
		float pos0[] = { 10.0f, 20.0f, -30.0f, 0.0f };
		float pos1[] = { -10.0f, -20.0f, 30.0f, 0.0f };
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, pos0, 0);
		gl.glLightfv(GL.GL_LIGHT1, GL.GL_POSITION, pos1, 0);

		// Draw the Graph
		displayGraph25D(drawable);

		if (takePicture) {
			int width = glCanvas.getWidth();
			int height = glCanvas.getHeight();
			int numPixels = width * height;
			picturePixelData = FloatBuffer.allocate(numPixels * 3);

			gl.glReadPixels(0, 0, width, height, GL.GL_RGB, GL.GL_FLOAT,
					picturePixelData);

			takePicture = false;
		}
	}

	/**
	 * Adjusts camera parameters when the canvas this view renders into is
	 * resized.
	 * <p>
	 * This method does not have to be called directly, since that is handled by
	 * the Java OpenGL system.
	 * 
	 * @see javax.media.opengl.GLEventListener#reshape(javax.media.opengl.GLAutoDrawable,
	 *      int, int, int, int)
	 */
	public void reshape(GLAutoDrawable drawable, int x, int y, int width,
			int height) {
		camera.setWidth(width);
		camera.setHeight(height);
	}

	/**
	 * Currently, this method is empty.
	 * <p>
	 * This method does not have to be called directly, since that is handled by
	 * the Java OpenGL system.
	 * 
	 * @see javax.media.opengl.GLEventListener#displayChanged(javax.media.opengl.GLAutoDrawable,
	 *      boolean, boolean)
	 */
	public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
			boolean deviceChanged) {
	}

	// ------------------------- DISPLAY GRAPH 25 D -------------------------

	/**
	 * Renders the <code>Graph25D</code> that is associated with this view
	 * into the given drawable.
	 * <p>
	 * For rendering individual <code>Graph2D</code>s, this method calls
	 * {@link #displayGraph2D displayGraph2D()}. Layers are rendered through a
	 * call to {@link #displayGraph2DLayer displayGraph2DLayer()}.
	 */
	protected void displayGraph25D(GLAutoDrawable drawable) {
		if (graph25D != null) {

			// 1st: DRAW LAYER EDGES
			for (LayerEdgeCursor lec = graph25D.layerEdges(); lec.ok(); lec
					.next()) {
				LayerEdgeRealizer25D ler = graph25D
						.getRealizer(lec.layerEdge());
				ler.paintGL(drawable, camera);
			}

			// 2nd: DRAW GRAPHS ( NODES, EDGES )

			for (GraphCursor gc = graph25D.graphs(); gc.ok(); gc.next()) {
				displayGraph2D(drawable, (Graph2D) gc.graph());
			}

			// 3rd: DRAW LAYERS
			// Since the Layers are transparent, they have to be drawn at the
			// end. Also, they have to be drawn in a back-to-front order, i.e.
			// those layers that are furthest apart from the camera need to be
			// drawn first, the closest ones need to be drawn last.

			// sort the layers according to their z-value, but keep the order
			// inside the graph25D intact
			Graph[] graphs = graph25D.getGraphArray();
			Arrays.sort(graphs, new Comparator() {
				public int compare(Object g1, Object g2) {
					double z1 = graph25D.getZ((Graph) g1);
					double z2 = graph25D.getZ((Graph) g2);
					if (z1 == z2)
						return 0;
					else
						return (z1 < z2) ? -1 : 1;
				}
			});

			// determine camera position
			double cameraZ = camera.getPosition().z;

			// draw all layers that are below the camera
			for (int i = 0; i < graphs.length; ++i) {
				if (graph25D.getZ(graphs[i]) > cameraZ)
					break;
				displayGraph2DLayer(drawable, (Graph2D) graphs[i]);
			}

			// draw all layers that are above the camera
			for (int i = graphs.length - 1; i >= 0; --i) {
				if (graph25D.getZ(graphs[i]) <= cameraZ)
					break;
				displayGraph2DLayer(drawable, (Graph2D) graphs[i]);
			}
		}
	}

	// ------------------------- DISPLAY GRAPH 2 D -------------------------

	/**
	 * Renders a <code>Graph2D</code> into the given drawable.
	 */
	protected void displayGraph2D(GLAutoDrawable drawable, Graph2D g2D) {
		GL gl = drawable.getGL();

		gl.glPushMatrix();

		// traslate origin to current layer
		gl.glTranslatef(0.0f, 0.0f, -(float) graph25D.getZ(g2D)); // translate
		// to
		// correct
		// z-coordiante

		// 1st: DRAW NODES
		for (NodeCursor nc = g2D.nodes(); nc.ok(); nc.next()) {
			NodeRealizer25D nr = (NodeRealizer25D) g2D.getRealizer(nc.node());
			nr.paintGL(drawable, camera);
		}

		// 2nd: DRAW EDGES
		for (EdgeCursor ec = g2D.edges(); ec.ok(); ec.next()) {
			Edge e = ec.edge();
			EdgeRealizer25D er = (EdgeRealizer25D) g2D.getRealizer(e);
			er.paintGL(drawable, camera);
		}

		gl.glPopMatrix();
	}

	/**
	 * Renders the layer representation of the given <code>Graph2D</code> into
	 * the given drawable.
	 */
	protected void displayGraph2DLayer(GLAutoDrawable drawable, Graph2D g2D) {
		GL gl = drawable.getGL();

		gl.glPushMatrix();

		// traslate origin to current layer
		gl.glTranslatef(0.0f, 0.0f, -(float) graph25D.getZ(g2D)); // translate
		// to
		// correct
		// z-coordiante

		// DRAW LAYER
		LayerRealizer25D lr = (LayerRealizer25D) graph25D.getRealizer(g2D);
		lr.paintGL(drawable, camera);

		gl.glPopMatrix();
	}

	/**
	 * Stops sending mouse events to all view modes that have been added to this
	 * graph.
	 */
	private void unregisterAllViewModes() {
		Iterator it = viewModes.iterator();
		while (it.hasNext()) {
			ViewMode25D vm = (ViewMode25D) it.next();
			glCanvas.removeMouseListener(vm);
			glCanvas.removeMouseMotionListener(vm);
			glCanvas.removeMouseWheelListener(vm);

		}
	}

	/**
	 * Adds the given view mode to this view and makes it active.
	 * <p>
	 * There can only be one active view mode at a time.
	 */
	public void addViewMode25D(ViewMode25D vm) {

		if (!viewModes.contains(vm)) {
			viewModes.add(vm);
			vm.setActiveView(this);
		}

		setActiveViewMode(vm);
	}

	/**
	 * Returns an interator to all view modes that have been added to this view.
	 */
	public Iterator getViewModes() {
		return viewModes.listIterator();
	}

	/**
	 * Sets the given view mode as the active view mode.
	 * <p>
	 * There can only be one active view mode at a time.
	 */
	public void setActiveViewMode(ViewMode25D vm) {

		unregisterAllViewModes();

		glCanvas.addMouseListener(vm);
		glCanvas.addMouseMotionListener(vm);
		glCanvas.addMouseWheelListener(vm);
	}

}

/*
 * Arrow a = Arrow.STANDARD; ArrowPainter.paintGL(drawable, a, new
 * Point3D(0,0,0), new Vector3D(-50,0,0), 20 );
 */

/*
 * Point3D x = new Point3D(0,0,0); Point3D y = new Point3D(100,0,0); Point3D z =
 * new Point3D(200,200,0);
 * 
 * Vector3D x2y = new Vector3D(x,y).normalize(); Vector3D y2z = new
 * Vector3D(y,z).normalize();
 * 
 * Vector3D xF = new Vector3D(1,0,0); Vector3D yF = x2y.getAdded(y2z); Vector3D
 * zF = new Vector3D(0,1,0); //Graphics3D.drawCylinder(drawable, x, y, 10.0f,
 * 10, true); //Graphics3D.drawCylinder(drawable, xF, x, y, yF, 10.0f, 10,
 * true); Graphics3D.drawCylinder(drawable, yF, y, z, zF, 10.0f, 10, true);
 */
