/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.view.realizer;

import java.awt.Color;

import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;

import y.base.Edge;
import y.geom.YPoint;
import y.view.EdgeRealizer;
import y.view.LineType;
import y.view.NodeRealizer;
import y.view.PolyLineEdgeRealizer;
import y25.graphics.BoundingBox;
import y25.graphics.Camera;
import y25.graphics.Graphics3D;
import y25.graphics.Point3D;
import y25.graphics.Vector3D;
import y25.view.EdgeRealizer25D;
import y25.view.NodeRealizer25D;

/**
 * Extends the <code>PolyLineEdgeRealizer</code> with 25D information.
 * <p>
 * This realizer aims to provide the same functionality as the
 * {@link y.view.PolyLineEdgeRealizer y.view.PolyLineEdgeRealizer}, except that
 * it can also be used inside a <code>Graph25D</code> and therefore rendered
 * into a <code>Graph25DView</code>.
 * <p>
 * When this realizer renders itself into an <code>Graph25DView</code>, it
 * does so by either rendering a simple line to visualize the edge or by
 * rendering a 3D cylinder. Whether a 2D or a 3D line is drawn can be set via
 * {@link #setLineDisplayMode setLineDisplayMode}. Note that the arrow head is
 * always rendered as a three dimensional cone.
 * 
 * @author Werner Jainek
 */
public class PolyLineEdgeRealizer25D extends PolyLineEdgeRealizer implements
		EdgeRealizer25D {

	/**
	 * When rendering a 3D cylinder, this determines the number of sides used.
	 */
	static int cylinderLineSides = 5;

	/**
	 * When rendering a 3D cylinder, this determines the number of bends used
	 * for self loops.
	 */
	static int cylinderLineSelfLoopBends = 10; // must be equal or greater than

	// 3

	/**
	 * When rendering a 2D line, this determines the number of bends used for
	 * self loops.
	 */
	static int lineSelfLoopBends = 20; // must be equal or greater than 3

	/**
	 * The length of the arrow head.
	 */
	static float arrowLength = 10;

	/**
	 * The radius of the arrow head's base.
	 */
	static float arrowRadius = 4;

	/**
	 * Additional scaling of the arrow head.
	 */
	static float arrowScale = 0;

	/**
	 * Current drawing mode for lines.
	 */
	protected int displayMode;

	/**
	 * Creates a new <code>PloyLineEdgeRealizer25D</code>. Edges willbe drawn
	 * as 2D lines.
	 */
	public PolyLineEdgeRealizer25D() {
		super();
		displayMode = MODE_2D;
	}

	/**
	 * Creates a new <code>PloyLineEdgeRealizer25D</code>, while trying to
	 * copy as many properties of the passed realizer as possible.
	 */
	public PolyLineEdgeRealizer25D(EdgeRealizer er) {
		super(er);

		if (er instanceof EdgeRealizer25D) {
			// member variable copies go here !
		}

		if (er instanceof PolyLineEdgeRealizer25D) {
			displayMode = ((PolyLineEdgeRealizer25D) er).getLineDisplayMode();
		}
	}

	/**
	 * Creates a copy of this realizer.
	 * 
	 * @see y.view.EdgeRealizer#createCopy(y.view.EdgeRealizer)
	 */
	public EdgeRealizer createCopy(EdgeRealizer er) {
		return new PolyLineEdgeRealizer25D(er);
	}

	/**
	 * Returns the current display mode for edge lines when drawing into an
	 * <code>Graph25DView</code>. Can be eiter
	 * <code>PolyLineEdgeRealizer.MODE_2D</code> or
	 * <code>PolyLineEdgeRealizer.MODE_3D</code>.
	 */
	public int getLineDisplayMode() {
		return displayMode;
	}

	/**
	 * Returns the current display mode for edge lines when drawing into an
	 * <code>Graph25DView</code>. <code>mode</code> can be eiter
	 * <code>PolyLineEdgeRealizer.MODE_2D</code> or
	 * <code>PolyLineEdgeRealizer.MODE_3D</code>.
	 */
	public void setLineDisplayMode(int mode) {
		displayMode = mode;
	}

	/**
	 * @see y25.view.EdgeRealizer25D#paintGL(javax.media.opengl.GLAutoDrawable,
	 *      y25.graphics.Camera)
	 */
	public void paintGL(GLAutoDrawable drawable, Camera camera) {

		// check if we have an attached edge
		Edge e = getEdge();
		if (e != null) {

			// set matarial
			setCylinderMaterial(drawable);

			// check for self-loops with only one bend
			if (e.isSelfLoop() && pointCount() == 1) {
				switch (displayMode) {
				default:
				case MODE_2D:
					paintGLLineSelfLoop(drawable, camera);
					break;
				case MODE_3D:
					paintGLCylinderLineSelfLoop(drawable, camera);
					break;
				}
			}

			else {
				switch (displayMode) {
				default:
				case MODE_2D:
					paintGLLine(drawable, camera);
					break;
				case MODE_3D:
					paintGLCylinderLine(drawable, camera);
					break;
				}
			}

			// unset material
			unsetCylinderMaterial(drawable);

		}
	}

	/**
	 * Renders the edge as a 2D line.
	 */
	protected void paintGLLine(GLAutoDrawable drawable, Camera camera) {
		GL gl = drawable.getGL();

		float lineWidth = (float) getLineRadius() * 2.0f;

		gl.glPushMatrix();

		// determine absolute node positions
		NodeRealizer snr = getSourceRealizer();
		NodeRealizer tnr = getTargetRealizer();
		Point3D source = new Point3D(
				(float) (snr.getCenterX() + getSourcePoint().x), (float) (snr
						.getCenterY() + getSourcePoint().y), 0);
		Point3D target = new Point3D(
				(float) (tnr.getCenterX() + getTargetPoint().x), (float) (tnr
						.getCenterY() + getTargetPoint().y), 0);

		// arrows: determine arrow tips and directions
		Vector3D sourceArrowTipToBase = new Vector3D(source, target);
		Point3D sourceArrowTip = ((NodeRealizer25D) snr).intersect(target,
				source, 0);

		Vector3D targetArrowTipToBase = new Vector3D(target, source);
		Point3D targetArrowTip = ((NodeRealizer25D) tnr).intersect(source,
				target, 0);

		gl.glLineWidth(lineWidth);
		gl.glBegin(GL.GL_LINE_STRIP);

		gl.glVertex3f(source.x, source.y, source.z);

		// iterate through all bends
		Point3D current = new Point3D(source);
		for (int i = 0; i < pointCount(); ++i) {

			// get coordiantes of next bend point
			YPoint pt = getPoint(i);
			Point3D next = new Point3D((float) pt.x, (float) pt.y, 0);
			gl.glVertex3f(next.x, next.y, 0);

			// arrows: find out where to draw them
			if (i == 0) {
				sourceArrowTip = null;
				targetArrowTip = null;
			}
			if (sourceArrowTip == null) {
				sourceArrowTip = ((NodeRealizer25D) snr).intersect(next,
						current, 0);
				sourceArrowTipToBase.set(current, next);
			}
			targetArrowTip = ((NodeRealizer25D) tnr)
					.intersect(current, next, 0);
			if (targetArrowTip != null) {
				targetArrowTipToBase.set(next, current);
			}

			// advance
			current = next;
		}

		gl.glVertex3f(target.x, target.y, target.z);

		gl.glEnd();

		// arrows: draw them
		targetArrowTip = ((NodeRealizer25D) tnr).intersect(current, target, 0);
		if (targetArrowTip != null) {
			targetArrowTipToBase.set(target, current);
		}
		if (sourceArrowTip != null)
			ArrowPainter.paintGL(drawable, camera, getSourceArrow(),
					sourceArrowTip, sourceArrowTipToBase.normalize().scale(
							arrowLength + lineWidth * arrowScale), arrowRadius);
		if (targetArrowTip != null)
			ArrowPainter.paintGL(drawable, camera, getTargetArrow(),
					targetArrowTip, targetArrowTipToBase.normalize().scale(
							arrowLength + lineWidth * arrowScale), arrowRadius);

		gl.glPopMatrix();
	}

	/**
	 * Renders the self loop edge as a 2D line.
	 */
	protected void paintGLLineSelfLoop(GLAutoDrawable drawable, Camera camera) {
		GL gl = drawable.getGL();

		float lineWidth = (float) getLineRadius() * 2.0f;

		gl.glPushMatrix();

		// get the coordinates of the node
		NodeRealizer snr = getSourceRealizer();
		Point3D node = new Point3D((float) snr.getCenterX(), (float) snr
				.getCenterY(), 0.0f);

		// get the coordinates of the only bend
		YPoint pt = getPoint(0);
		Point3D bend = new Point3D((float) pt.x, (float) pt.y, 0);

		// determine the center of the self loop circle
		Point3D center = Point3D.lerp(node, bend, 0.5f);

		// horizontal-axis points from center to node
		Vector3D axisH = new Vector3D(center, node);

		// vertical-axis is perpendicular to axisH and to z-axis
		Vector3D axisV = new Vector3D(0, 0, -1).cross(axisH);

		// angle steps
		float angleStep = (float) (2.0 * Math.PI) / (float) lineSelfLoopBends;
		float angle = -angleStep;

		// arrows: determine arrow tips and directions
		Vector3D sourceArrowTipToBase = new Vector3D();
		Point3D sourceArrowTip = null;

		Vector3D targetArrowTipToBase = new Vector3D();
		Point3D targetArrowTip = null;

		// create the circle
		gl.glLineWidth(lineWidth);
		gl.glBegin(GL.GL_LINE_STRIP);

		Point3D current = new Point3D(node);
		gl.glVertex3f(current.x, current.y, current.z);

		for (int i = 0; i < lineSelfLoopBends; ++i) {

			// calculate position of next end point
			Point3D next = new Point3D(center);
			next.translate(axisH.getScaled((float) Math.cos(angle)));
			next.translate(axisV.getScaled((float) Math.sin(angle)));

			// draw the line segment
			gl.glVertex3f((float) next.x, (float) next.y, 0);

			// arrows: find out where to draw them
			if (i == 0) {
				sourceArrowTip = null;
				targetArrowTip = null;
			}
			if (sourceArrowTip == null) {
				sourceArrowTip = ((NodeRealizer25D) snr).intersect(next,
						current, 0);
				sourceArrowTipToBase.set(current, next);
			}
			targetArrowTip = ((NodeRealizer25D) snr)
					.intersect(current, next, 0);
			if (targetArrowTip != null) {
				targetArrowTipToBase.set(next, current);
			}

			// advance
			current = next;
			angle -= angleStep;
		}

		gl.glEnd();

		// arrows: draw them
		targetArrowTip = ((NodeRealizer25D) snr).intersect(current, node, 0);
		if (targetArrowTip != null) {
			targetArrowTipToBase.set(node, current);
		}
		if (sourceArrowTip != null)
			ArrowPainter.paintGL(drawable, camera, getSourceArrow(),
					sourceArrowTip, sourceArrowTipToBase.normalize().scale(
							arrowLength + lineWidth * arrowScale), arrowRadius);
		if (targetArrowTip != null)
			ArrowPainter.paintGL(drawable, camera, getTargetArrow(),
					targetArrowTip, targetArrowTipToBase.normalize().scale(
							arrowLength + lineWidth * arrowScale), arrowRadius);

		gl.glPopMatrix();
	}

	/**
	 * Sets the cylinder material.
	 */
	protected void setCylinderMaterial(GLAutoDrawable drawable) {
		GL gl = drawable.getGL();

		gl.glDisable(GL.GL_COLOR_MATERIAL);
		Color nodeColor = getLineColor();
		float mat_none[] = { 0.0f, 0.0f, 0.0f, 1.0f };
		float mat_diffuse[] = { ((float) nodeColor.getRed()) / 255.0f,
				((float) nodeColor.getGreen()) / 255.0f,
				((float) nodeColor.getBlue()) / 255.0f, 1.0f };
		float mat_noshininess[] = { 0.0f };
		gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE, mat_diffuse, 0);
		gl.glMaterialfv(GL.GL_FRONT, GL.GL_SPECULAR, mat_none, 0);
		gl.glMaterialfv(GL.GL_FRONT, GL.GL_SHININESS, mat_noshininess, 0);
		gl.glMaterialfv(GL.GL_FRONT, GL.GL_EMISSION, mat_none, 0);
	}

	/**
	 * Unsets the cylinder material.
	 */
	protected void unsetCylinderMaterial(GLAutoDrawable drawable) {
		GL gl = drawable.getGL();
		gl.glEnable(GL.GL_COLOR_MATERIAL);
	}

	/**
	 * Renders the edge as a 3D cylinder.
	 */
	protected void paintGLCylinderLine(GLAutoDrawable drawable, Camera camera) {

		GL gl = drawable.getGL();

		float radius = (float) getLineRadius();

		gl.glPushMatrix();

		// determine absolute node positions
		NodeRealizer snr = getSourceRealizer();
		NodeRealizer tnr = getTargetRealizer();
		Point3D source = new Point3D(
				(float) (snr.getCenterX() + getSourcePoint().x), (float) (snr
						.getCenterY() + getSourcePoint().y), 0);
		Point3D target = new Point3D(
				(float) (tnr.getCenterX() + getTargetPoint().x), (float) (tnr
						.getCenterY() + getTargetPoint().y), 0);

		// define the normals of the planes that cut the cylinder
		Vector3D sourceTiltPlane = new Vector3D(source, target);
		Vector3D targetTiltPlane = sourceTiltPlane;

		// arrows: determine arrow tips and directions
		Vector3D sourceArrowTipToBase = new Vector3D(source, target);
		Point3D sourceArrowTip = ((NodeRealizer25D) snr).intersect(target,
				source, 0);

		Vector3D targetArrowTipToBase = new Vector3D(target, source);
		Point3D targetArrowTip = ((NodeRealizer25D) tnr).intersect(source,
				target, 0);

		// iterate through all bends
		Point3D current = source;
		Vector3D currentTiltPlane = sourceTiltPlane;
		for (int i = 0; i < pointCount(); ++i) {

			// get coordiantes of next bend point
			YPoint pt = getPoint(i);
			Point3D next = new Point3D((float) pt.x, (float) pt.y, 0);

			// get coordinates of second next bend point (or of target node, if
			// no other bend points exist
			Point3D nextnext;
			if (i < pointCount() - 1) {
				YPoint tmppt = getPoint(i + 1);
				nextnext = new Point3D((float) tmppt.x, (float) tmppt.y, 0);
			} else {
				nextnext = target;
			}

			// if we have any bends, adjust the source and target tiltplanes
			if (i == 0) { // we are at the first bend
				currentTiltPlane = new Vector3D(source, next);
			}
			if (i == pointCount() - 1) { // we are at the last bend
				targetTiltPlane = new Vector3D(next, target);
			}

			// compute the normal of the tilt plane for point 'next'
			Vector3D current2next = new Vector3D(current, next).normalize();
			Vector3D next2nextnext = new Vector3D(next, nextnext).normalize();
			Vector3D nextTiltPlane = current2next.getAdded(next2nextnext);

			// if the line segments (current-next) and (next-nextnext) are
			// parallel, we have to manually set the normal
			if (nextTiltPlane.isZero()) {
				nextTiltPlane.set(current2next);
			}

			// if line segments form a too narrow bend, draw a capped cylinder
			boolean capped = false;
			if (Math.abs(nextTiltPlane.getNormalized().dot(current2next)) < 0.2) {
				nextTiltPlane.set(current2next);
				capped = true;
			}

			// arrows: find out where to draw them
			if (i == 0) {
				sourceArrowTip = null;
				targetArrowTip = null;
			}
			if (sourceArrowTip == null) {
				sourceArrowTip = ((NodeRealizer25D) snr).intersect(next,
						current, 0);
				sourceArrowTipToBase.set(current, next);
			}
			targetArrowTip = ((NodeRealizer25D) tnr)
					.intersect(current, next, 0);
			if (targetArrowTip != null) {
				targetArrowTipToBase.set(next, current);
			}

			// draw line segment
			Graphics3D.drawCylinder(drawable, currentTiltPlane, current, next,
					nextTiltPlane, radius, cylinderLineSides, capped);

			// advance
			current = next;
			currentTiltPlane = nextTiltPlane;
		}

		// draw last line segment
		Graphics3D.drawCylinder(drawable, currentTiltPlane, current, target,
				targetTiltPlane, radius, cylinderLineSides, false);
		// Graphics3D.drawCylinder(drawable, src, target, radius, sides);

		// arrows: draw them
		targetArrowTip = ((NodeRealizer25D) tnr).intersect(current, target, 0);
		if (targetArrowTip != null) {
			targetArrowTipToBase.set(target, current);
		}
		if (sourceArrowTip != null)
			ArrowPainter
					.paintGL(drawable, camera, getSourceArrow(),
							sourceArrowTip, sourceArrowTipToBase.normalize()
									.scale(
											arrowLength + radius * 2
													* arrowScale), arrowRadius);
		if (targetArrowTip != null)
			ArrowPainter
					.paintGL(drawable, camera, getTargetArrow(),
							targetArrowTip, targetArrowTipToBase.normalize()
									.scale(
											arrowLength + radius * 2
													* arrowScale), arrowRadius);

		gl.glPopMatrix();
	}

	/**
	 * Renders the self loop edge as a 3D cylinder.
	 */
	protected void paintGLCylinderLineSelfLoop(GLAutoDrawable drawable,
			Camera camera) {

		GL gl = drawable.getGL();

		float radius = (float) getLineRadius();

		gl.glPushMatrix();

		// get the coordinates of the node
		NodeRealizer snr = getSourceRealizer();
		Point3D node = new Point3D((float) snr.getCenterX(), (float) snr
				.getCenterY(), 0.0f);

		// get the coordinates of the only bend
		YPoint pt = getPoint(0);
		Point3D bend = new Point3D((float) pt.x, (float) pt.y, 0);

		// determine the center of the self loop circle
		Point3D center = Point3D.lerp(node, bend, 0.5f);

		// horizontal-axis points from center to node
		Vector3D axisH = new Vector3D(center, node);

		// vertical-axis is perpendicular to axisH and to z-axis
		Vector3D axisV = new Vector3D(0, 0, -1).cross(axisH);

		// arrows: determine arrow tips and directions
		Vector3D sourceArrowTipToBase = new Vector3D();
		Point3D sourceArrowTip = null;

		Vector3D targetArrowTipToBase = new Vector3D();
		Point3D targetArrowTip = null;

		// create the circle
		Point3D current = new Point3D(node);
		Vector3D currentTiltPlane = new Vector3D(axisV);
		float angleStep = (float) (2.0 * Math.PI)
				/ (float) cylinderLineSelfLoopBends;
		float angle = -angleStep;

		for (int i = 0; i < cylinderLineSelfLoopBends; ++i) {

			// calculate position of next end point
			Point3D next = new Point3D(center);
			next.translate(axisH.getScaled((float) Math.cos(angle)));
			next.translate(axisV.getScaled((float) Math.sin(angle)));

			// calculate normal of next end point
			Vector3D nextTiltPlane = new Vector3D(0, 0, 0);
			nextTiltPlane.add(axisH.getScaled(-(float) Math.sin(angle)));
			nextTiltPlane.add(axisV.getScaled((float) Math.cos(angle)));

			// draw the line segment
			Graphics3D.drawCylinder(drawable, currentTiltPlane, current, next,
					nextTiltPlane, radius, cylinderLineSides, false);

			// arrows: find out where to draw them
			if (i == 0) {
				sourceArrowTip = null;
				targetArrowTip = null;
			}
			if (sourceArrowTip == null) {
				sourceArrowTip = ((NodeRealizer25D) snr).intersect(next,
						current, 0);
				sourceArrowTipToBase.set(current, next);
			}
			targetArrowTip = ((NodeRealizer25D) snr)
					.intersect(current, next, 0);
			if (targetArrowTip != null) {
				targetArrowTipToBase.set(next, current);
			}

			// advance
			current = next;
			currentTiltPlane = nextTiltPlane;
			angle -= angleStep;
		}

		// arrows: draw them
		targetArrowTip = ((NodeRealizer25D) snr).intersect(current, node, 0);
		if (targetArrowTip != null) {
			targetArrowTipToBase.set(node, current);
		}
		if (sourceArrowTip != null)
			ArrowPainter
					.paintGL(drawable, camera, getSourceArrow(),
							sourceArrowTip, sourceArrowTipToBase.normalize()
									.scale(
											arrowLength + radius * 2
													* arrowScale), arrowRadius);
		if (targetArrowTip != null)
			ArrowPainter
					.paintGL(drawable, camera, getTargetArrow(),
							targetArrowTip, targetArrowTipToBase.normalize()
									.scale(
											arrowLength + radius * 2
													* arrowScale), arrowRadius);

		gl.glPopMatrix();
	}

	/**
	 * @see y25.layout.EdgeLayout25D#getRelativeBoundingBox()
	 */
	public BoundingBox getRelativeBoundingBox() {

		BoundingBox bb = new BoundingBox();

		bb.includeZ(0);

		// if we have an assigned edge, then include its endpoints in the
		// bounding box
		Edge e = getEdge();
		if (e != null) {
			bb.includeX((float) getSourceRealizer().getX());
			bb.includeY((float) getSourceRealizer().getY());
			bb.includeX((float) getTargetRealizer().getX());
			bb.includeY((float) getTargetRealizer().getY());
		}

		// now include all bend points
		for (int i = 0; i < pointCount(); ++i) {
			YPoint pt = getPoint(i);
			bb.includeX((float) pt.x);
			bb.includeY((float) pt.y);
		}

		// grow the bounding box by the radius of the line
		bb.addBorder((float) getLineRadius());

		return bb;
	}

	/**
	 * Returns the radius of the edge based on the
	 * {@link y.view.LineType LineType} of this realizer.
	 */
	public double getLineRadius() {
		double radius = 1.0;
		LineType lineType = getLineType();
		if (lineType.equals(LineType.LINE_1)) {
			radius = 1.0;
		} else if (lineType.equals(LineType.LINE_2)) {
			radius = 2.0;
		} else if (lineType.equals(LineType.LINE_3)) {
			radius = 3.0;
		} else if (lineType.equals(LineType.LINE_4)) {
			radius = 4.0;
		} else if (lineType.equals(LineType.LINE_5)) {
			radius = 5.0;
		} else if (lineType.equals(LineType.LINE_6)) {
			radius = 6.0;
		} else if (lineType.equals(LineType.LINE_7)) {
			radius = 7.0;
		} else if (lineType.equals(LineType.DOTTED_1)) {
			radius = 1.0;
		} else if (lineType.equals(LineType.DOTTED_2)) {
			radius = 2.0;
		} else if (lineType.equals(LineType.DOTTED_3)) {
			radius = 3.0;
		} else if (lineType.equals(LineType.DOTTED_4)) {
			radius = 4.0;
		} else if (lineType.equals(LineType.DOTTED_5)) {
			radius = 5.0;
		} else if (lineType.equals(LineType.DASHED_1)) {
			radius = 1.0;
		} else if (lineType.equals(LineType.DASHED_2)) {
			radius = 2.0;
		} else if (lineType.equals(LineType.DASHED_3)) {
			radius = 3.0;
		} else if (lineType.equals(LineType.DASHED_4)) {
			radius = 4.0;
		} else if (lineType.equals(LineType.DASHED_5)) {
			radius = 5.0;
		}

		return radius / 2.0;
	}
}
