/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.view.realizer;

import java.awt.Color;

import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;

import y25.graphics.BoundingBox;
import y25.graphics.Camera;
import y25.graphics.Point3D;
import y25.view.LayerLabel;
import y25.view.LayerRealizer25D;

import com.sun.opengl.util.GLUT;

/**
 * Displays a layer as a transparent rectangle.
 * 
 * @author Werner Jainek
 * 
 */
public class TransparentLayerRealizer25D implements LayerRealizer25D {

	/**
	 * The z-position of the layer.
	 */
	protected double z;

	/**
	 * The relative bounding box of the layer.
	 */
	BoundingBox relativeBB;

	/**
	 * The red component of the layer rectangle.
	 */
	float r;

	/**
	 * The green component of the layer rectangle.
	 */
	float g;

	/**
	 * The blue component of the layer rectangle.
	 */
	float b;

	/**
	 * The opacity fot the layer rectangle.
	 */
	float opacity;

	/**
	 * Stores the selection state for the layer.
	 */
	boolean selected;

	/**
	 * The label of this realizer.
	 */
	LayerLabel label;

	/**
	 * Creates a new <code>TransparentLayerRealizer25D</code> with zero size.
	 */
	public TransparentLayerRealizer25D() {
		label = new LayerLabel();
		relativeBB = new BoundingBox();
		setOpacity(0.2f);
		setRGB(250, 250, 250);
	}

	/**
	 * @see y25.view.LayerRealizer25D#createCopy()
	 */
	public LayerRealizer25D createCopy() {
		TransparentLayerRealizer25D lr = new TransparentLayerRealizer25D();
		lr.setZ(z);
		lr.setRelativeBoundingBox(relativeBB);
		lr.r = r;
		lr.g = g;
		lr.b = b;
		lr.setOpacity(opacity);
		lr.setSelected(selected);
		lr.setLabelText(getLabelText());
		return lr;
	}

	/**
	 * @see y25.layout.LayerLayout25D#getZ()
	 */
	public double getZ() {
		return z;
	}

	/**
	 * @see y25.layout.LayerLayout25D#setZ(double)
	 */
	public void setZ(double z) {
		this.z = z;
	}

	/**
	 * @see y25.layout.LayerLayout25D#setRelativeBoundingBox(y25.graphics.BoundingBox)
	 */
	public void setRelativeBoundingBox(BoundingBox bb) {
		relativeBB.set(bb);
	}

	/**
	 * @see y25.layout.LayerLayout25D#getRealativeBoundingBox()
	 */
	public BoundingBox getRealativeBoundingBox() {
		return new BoundingBox(relativeBB);
	}

	/**
	 * Sets the color of the layer rectangle to the specified red, green and
	 * blue values. These values have to be in the range <code>[0,255]</code>.
	 */
	public void setRGB(int r, int g, int b) {
		this.r = ((float) Math.min(Math.max(r, 0), 255)) / 255.0f;
		this.g = ((float) Math.min(Math.max(g, 0), 255)) / 255.0f;
		this.b = ((float) Math.min(Math.max(b, 0), 255)) / 255.0f;
	}

	/**
	 * Sets the opacity of the layer rectangle to the passed value. The value
	 * has to be in the range [0.0,1.0].
	 */
	public void setOpacity(float o) {
		opacity = Math.min(Math.max(o, 0.0f), 1.0f);
	}

	/**
	 * Returns the opacity of the layer rectangle.
	 */
	public float getOpacity() {
		return opacity;
	}

	/**
	 * @see y25.view.LayerRealizer25D#paintGL(javax.media.opengl.GLAutoDrawable,
	 *      y25.graphics.Camera)
	 */
	public void paintGL(GLAutoDrawable drawable, Camera camera) {
		GL gl = drawable.getGL();

		// paint labels first
		paintLabels(drawable, camera);

		gl.glPushAttrib(GL.GL_CURRENT_BIT);

		Point3D min = relativeBB.getMin();
		Point3D max = relativeBB.getMax();

		// draw border
		if (opacity > 0.01 || selected) {
			float op = (opacity + 0.2f >= 1.0f) ? 1.0f : opacity + 0.2f;
			if (selected) {
				gl.glLineWidth(3.0f);
				gl.glColor4f(Math.max(r - 0.4f, 0.0f),
						Math.max(g - 0.4f, 0.0f), Math.max(b - 0.4f, 0.0f), op);
			} else {
				gl.glLineWidth(2.0f);
				gl.glColor4f(r, g, b, op);
			}

			gl.glBegin(GL.GL_LINE_STRIP);
			gl.glVertex3f(min.x, min.y, 0);
			gl.glVertex3f(min.x, max.y, 0);
			gl.glVertex3f(max.x, max.y, 0);
			gl.glVertex3f(max.x, min.y, 0);
			gl.glVertex3f(min.x, min.y, 0);
			gl.glEnd();
		}

		// draw the plane
		gl.glBegin(GL.GL_QUADS);

		gl.glNormal3f(0, 0, -1);
		gl.glColor4f(r, g, b, opacity);

		// upper side
		gl.glVertex3f(min.x, min.y, 0);
		gl.glVertex3f(min.x, max.y, 0);
		gl.glVertex3f(max.x, max.y, 0);
		gl.glVertex3f(max.x, min.y, 0);

		// lowerSide
		gl.glVertex3f(max.x, min.y, 0);
		gl.glVertex3f(max.x, max.y, 0);
		gl.glVertex3f(min.x, max.y, 0);
		gl.glVertex3f(min.x, min.y, 0);

		gl.glEnd();

		gl.glPopAttrib();
	}

	/**
	 * Renders the labels of this realizer into the given drawable.
	 * 
	 * @param drawable
	 *            the drawable that the labels will be rendered into.
	 * @param camera
	 *            used to determine the orientation of the label such that is
	 *            faces the viewer.
	 */
	public void paintLabels(GLAutoDrawable drawable, Camera camera) {
		GL gl = drawable.getGL();
		GLUT glut = new GLUT();

		// 1. find correct corner to position the label
		// /////////////////////////////////////////////

		Point3D min = relativeBB.getMin();
		Point3D max = relativeBB.getMax();

		Point3D corners[] = { new Point3D(min.x, min.y, 0),
				new Point3D(min.x, max.y, 0), new Point3D(max.x, max.y, 0),
				new Point3D(max.x, min.y, 0) };

		int rightmostCornerIndex = 0;
		Point3D rightmostCorner = camera
				.getScreenCoordinates(corners[rightmostCornerIndex]);
		for (int i = 1; i < corners.length; ++i) {
			Point3D currentCorner = camera.getScreenCoordinates(corners[i]);
			if (currentCorner.x > rightmostCorner.x) {
				rightmostCorner = currentCorner;
				rightmostCornerIndex = i;
			}
		}
		Point3D labelPos = corners[rightmostCornerIndex];

		// 2. draw the label
		// /////////////////////////////////////////////

		// retrieve raster coordinates of position
		Point3D rasterPos = camera.getHUDCoordinates(drawable, labelPos);

		if (rasterPos != null) {

			// Leads to same x/y coordinates, but z-values not correct as they
			// depend on device
			// Point3D rasterPos2 = camera.getScreenCoordinates(pos);
			// rasterPos.x = rasterPos2.x;
			// rasterPos.y = rasterPos2.y;

			// activate "head-up-display"
			camera.activateHUD(drawable);

			// specify the material
			gl.glDisable(GL.GL_COLOR_MATERIAL);
			Color textColor = new Color(0, 0, 0);
			float mat_none[] = { 0.0f, 0.0f, 0.0f, 1.0f };
			float mat_diffuse[] = { ((float) textColor.getRed()) / 255.0f,
					((float) textColor.getGreen()) / 255.0f,
					((float) textColor.getBlue()) / 255.0f, 1.0f };
			float mat_noshininess[] = { 0.0f };
			gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE,
					mat_diffuse, 0);
			gl.glMaterialfv(GL.GL_FRONT, GL.GL_SPECULAR, mat_none, 0);
			gl.glMaterialfv(GL.GL_FRONT, GL.GL_SHININESS, mat_noshininess, 0);
			gl.glMaterialfv(GL.GL_FRONT, GL.GL_EMISSION, mat_none, 0);

			// draw the text
			String labelText = label.getText();

			int font = GLUT.BITMAP_HELVETICA_12;
			gl.glRasterPos3f(rasterPos.x + 12, rasterPos.y - 4, rasterPos.z);
			glut.glutBitmapString(font, labelText);

			gl.glEnable(GL.GL_COLOR_MATERIAL);

			camera.deactivateHUD(drawable);
		}

	}

	/**
	 * @see y25.view.LayerRealizer25D#isSelected()
	 */
	public boolean isSelected() {
		return selected;
	}

	/**
	 * @see y25.view.LayerRealizer25D#setSelected(boolean)
	 */
	public void setSelected(boolean selected) {
		this.selected = selected;
	}

	/**
	 * @see y25.view.LayerRealizer25D#getLabelText()
	 */
	public String getLabelText() {
		return label.getText();
	}

	public void setLabelText(String text) {
		label.setText(text);
	}

}
